from .constants import (
    Classref, Constant, Double, Fieldref, Float, Integer, InterfaceMethodref,
    Long, Methodref, String,
)


# From: https://en.wikipedia.org/wiki/Java_bytecode_instruction_listings
# Reference" http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-6.html#jvms-6.2

##########################################################################
# 6.2. Opcodes
##########################################################################


class Opcode:
    opcodes = None

    def __init__(self):
        self.references = []
        self.starts_line = None

    def __repr__(self):
        return '<%s%s>' % (self.__class__.__name__, self.__arg_repr__())

    def __arg_repr__(self):
        return ''

    def __len__(self):
        return 1

    @classmethod
    def read(cls, reader, dump=None):
        code = reader.read_u1()
        # Create an index of all known opcodes.
        if Opcode.opcodes is None:
            Opcode.opcodes = {}
            for name in globals():
                klass = globals()[name]
                try:
                    if name != 'Opcode' and issubclass(klass, Opcode):
                        Opcode.opcodes[klass.code] = klass
                except TypeError:
                    pass
        instance = Opcode.opcodes[code].read_extra(reader, dump)
        if dump:
            reader.debug("    " * dump, '%3d: %s' % (reader.offset, instance))
        return instance

    @classmethod
    def read_extra(cls, reader, dump=None):
        return cls()

    def write(self, writer):
        writer.write_u1(self.code)
        self.write_extra(writer)

    def write_extra(self, writer):
        pass

    def resolve(self, constant_pool):
        pass

    @property
    def stack_effect(self):
        return self.produce_count - self.consume_count

    def process(self, context):
        return True


class AALOAD(Opcode):
    # Load onto the stack a reference from an array
    # Stack: arrayref, index → value
    code = 0x32

    def __init__(self):
        super(AALOAD, self).__init__()

    @property
    def consume_count(self):
        return 2

    @property
    def produce_count(self):
        return 1


class AASTORE(Opcode):
    # Store into a reference in an array
    # Stack: arrayref, index, value →
    code = 0x53

    def __init__(self):
        super(AASTORE, self).__init__()

    @property
    def consume_count(self):
        return 3

    @property
    def produce_count(self):
        return 0


class ACONST_NULL(Opcode):
    # Push a null reference onto the stack
    # Stack: → null
    code = 0x01

    def __init__(self):
        super(ACONST_NULL, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ALOAD(Opcode):
    # Load a reference onto the stack from a local variable #index
    # Args(1): index
    # Stack: → objectref
    code = 0x19

    def __init__(self, var):
        super(ALOAD, self).__init__()
        self.var = var

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.var

    @classmethod
    def read_extra(cls, reader, dump=None):
        var = reader.read_u1()
        return cls(var)

    def write_extra(self, writer):
        writer.write_u1(self.var)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ALOAD_0(Opcode):
    # Load a reference onto the stack from local variable 0
    # Stack: → objectref
    code = 0x2a

    def __init__(self):
        super(ALOAD_0, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ALOAD_1(Opcode):
    # Load a reference onto the stack from local variable 1
    # Stack: → objectref
    code = 0x2b

    def __init__(self):
        super(ALOAD_1, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ALOAD_2(Opcode):
    # Load a reference onto the stack from local variable 2
    # Stack: → objectref
    code = 0x2c

    def __init__(self):
        super(ALOAD_2, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ALOAD_3(Opcode):
    # Load a reference onto the stack from local variable 3
    # Stack: → objectref
    code = 0x2d

    def __init__(self):
        super(ALOAD_3, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ANEWARRAY(Opcode):
    # Create a new array of references of length count and component type identified
    # by the class reference index (indexbyte1 << 8 + indexbyte2) in the constant
    # pool
    # Args(2): indexbyte1, indexbyte2
    # Stack: count → arrayref
    code = 0xbd

    def __init__(self, class_name):
        super(ANEWARRAY, self).__init__()
        self.klass = Classref(class_name)

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.klass.name

    @classmethod
    def read_extra(cls, reader, dump=None):
        klass = reader.read_u2()
        return cls(reader.constant_pool[klass].name.bytes.decode('mutf-8'))

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.klass))

    def resolve(self, constant_pool):
        self.klass.resolve(constant_pool)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class ARETURN(Opcode):
    # Return a reference from a method
    # Stack: objectref → [empty]
    code = 0xb0

    def __init__(self):
        super(ARETURN, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ARRAYLENGTH(Opcode):
    # Get the length of an array
    # arrayref → length
    code = 0xbe

    def __init__(self):
        super(ARRAYLENGTH, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class ASTORE(Opcode):
    # Store a reference into a local variable #index
    # Args(1): index
    # Stack: objectref →
    code = 0x3a

    def __init__(self, var):
        super(ASTORE, self).__init__()
        self.var = var

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.var

    @classmethod
    def read_extra(cls, reader, dump=None):
        var = reader.read_u1()
        return cls(var)

    def write_extra(self, writer):
        writer.write_u1(self.var)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ASTORE_0(Opcode):
    # Store a reference into local variable 0
    # Stack: objectref →
    code = 0x4b

    def __init__(self):
        super(ASTORE_0, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ASTORE_1(Opcode):
    # Store a reference into local variable 1
    # Stack: objectref →
    code = 0x4c

    def __init__(self):
        super(ASTORE_1, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ASTORE_2(Opcode):
    # Store a reference into local variable 2
    # Stack: objectref →
    code = 0x4d

    def __init__(self):
        super(ASTORE_2, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ASTORE_3(Opcode):
    # Store a reference into local variable 3
    # Stack: objectref →
    code = 0x4e

    def __init__(self):
        super(ASTORE_3, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ATHROW(Opcode):
    # Throws an error or exception (notice that the rest of the stack is cleared,
    # leaving only a reference to the Throwable)
    # Stack: objectref → [empty], objectref
    code = 0xbf

    def __init__(self):
        super(ATHROW, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 0


class BALOAD(Opcode):
    # Load a byte or Boolean value from an array
    # Stack: arrayref, index → value
    code = 0x33

    def __init__(self):
        super(BALOAD, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 3


class BASTORE(Opcode):
    # Store a byte or Boolean value into an array
    # Stack: arrayref, index, value →
    code = 0x54

    def __init__(self):
        super(BASTORE, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 3


class BIPUSH(Opcode):
    # Push a byte onto the stack as an integer value
    # Args(1) byte
    # Stack: → value
    code = 0x10

    def __init__(self, const):
        super(BIPUSH, self).__init__()
        self.const = const

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' ' + repr(self.const)

    @classmethod
    def read_extra(cls, reader, dump=None):
        const = reader.read_u1()
        return cls(const)

    def write_extra(self, writer):
        writer.write_u1(self.const)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class BREAKPOINT(Opcode):
    # Reserved for breakpoints in Java debuggers; should not appear in any class file
    code = 0xca

    def __init__(self):
        super(BREAKPOINT, self).__init__()


class CALOAD(Opcode):
    # Load a char from an array
    # Stack: arrayref, index → value
    code = 0x34

    def __init__(self):
        super(CALOAD, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class CASTORE(Opcode):
    # Store a char into an array
    # Stack: arrayref, index, value →
    code = 0x55

    def __init__(self):
        super(CASTORE, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 3


class CHECKCAST(Opcode):
    # Checks whether an objectref is of a certain type, the class reference of which
    # is in the constant pool at index (indexbyte1 << 8 + indexbyte2)
    # Args(2): indexbyte1, indexbyte2
    # Stack: objectref → objectref
    code = 0xc0

    def __init__(self, class_name):
        super(CHECKCAST, self).__init__()
        self.klass = Classref(class_name)

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % (self.klass)

    @classmethod
    def read_extra(cls, reader, dump=None):
        class_name = reader.constant_pool[reader.read_u2()].name.bytes.decode('mutf-8')
        return cls(class_name)

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.klass))

    def resolve(self, constant_pool):
        self.klass.resolve(constant_pool)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class D2F(Opcode):
    # Convert a double to a float
    # Stack: value → result
    code = 0x90

    def __init__(self):
        super(D2F, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class D2I(Opcode):
    # Convert a double to an int
    # Stack: value → result
    code = 0x8e

    def __init__(self):
        super(D2I, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class D2L(Opcode):
    # Convert a double to a long
    # Stack: value → result
    code = 0x8f

    def __init__(self):
        super(D2L, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class DADD(Opcode):
    # Add two doubles
    # Stack: value1, value2 → result
    code = 0x63

    def __init__(self):
        super(DADD, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class DALOAD(Opcode):
    # Load a double from an array
    # Stack: arrayref, index → value
    code = 0x31

    def __init__(self):
        super(DALOAD, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class DASTORE(Opcode):
    # Store a double into an array
    # Stack: arrayref, index, value →
    code = 0x52

    def __init__(self):
        super(DASTORE, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 3


class DCMPG(Opcode):
    # Compare two doubles
    # Stack: value1, value2 → result
    code = 0x98

    def __init__(self):
        super(DCMPG, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class DCMPL(Opcode):
    # Compare two doubles
    # Stack: value1, value2 → result
    code = 0x97

    def __init__(self):
        super(DCMPL, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class DCONST_0(Opcode):
    # Push the constant 0.0 onto the stack
    # Stack: → 0.0
    code = 0x0e

    def __init__(self):
        super(DCONST_0, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class DCONST_1(Opcode):
    # Push the constant 1.0 onto the stack
    # Stack:  → 1.0
    code = 0x0f

    def __init__(self):
        super(DCONST_1, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class DDIV(Opcode):
    # Divide two doubles
    # Stack: value1, value2 → result
    code = 0x6f

    def __init__(self):
        super(DDIV, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class DLOAD(Opcode):
    # Load a double value from a local variable #index
    # Args(1): index
    # Stack: → value
    code = 0x18

    def __init__(self, var):
        super(DLOAD, self).__init__()
        self.var = var

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.var

    @classmethod
    def read_extra(cls, reader, dump=None):
        var = reader.read_u1()
        return cls(var)

    def write_extra(self, writer):
        writer.write_u1(self.var)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class DLOAD_0(Opcode):
    # Load a double from local variable 0
    # Stack: → value
    code = 0x26

    def __init__(self, var):
        super(DLOAD_0, self).__init__()
        self.var = var

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.var

    @classmethod
    def read_extra(cls, reader, dump=None):
        var = reader.read_u1()
        return cls(var)

    def write_extra(self, writer):
        writer.write_u1(self.var)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class DLOAD_1(Opcode):
    # Load a double from local variable 1
    # Stack: → value
    code = 0x27

    def __init__(self):
        super(DLOAD_1, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class DLOAD_2(Opcode):
    # Load a double from local variable 2
    # Stack: → value
    code = 0x28

    def __init__(self):
        super(DLOAD_2, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class DLOAD_3(Opcode):
    # Load a double from local variable 3
    # Stack: → value
    code = 0x29

    def __init__(self):
        super(DLOAD_3, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class DMUL(Opcode):
    # Multiply two doubles
    # Stack: value1, value2 → result
    code = 0x6b

    def __init__(self):
        super(DMUL, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class DNEG(Opcode):
    # Negate a double
    # Stack: value → result
    code = 0x77

    def __init__(self):
        super(DNEG, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class DREM(Opcode):
    # Get the remainder from a division between two doubles
    # value1, value2 → result
    code = 0x73

    def __init__(self):
        super(DREM, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class DRETURN(Opcode):
    # Return a double from a method
    # Stack:value → [empty]
    code = 0xaf

    def __init__(self):
        super(DRETURN, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class DSTORE(Opcode):
    # Store a double value into a local variable #index
    # Args(1): index
    # Stack: value →
    code = 0x39

    def __init__(self, var):
        super(DSTORE, self).__init__()
        self.var = var

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.var

    @classmethod
    def read_extra(cls, reader, dump=None):
        var = reader.read_u1()
        return cls(var)

    def write_extra(self, writer):
        writer.write_u1(self.var)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class DSTORE_0(Opcode):
    # Store a double into local variable 0
    # Stack: value →
    code = 0x47

    def __init__(self):
        super(DSTORE_0, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class DSTORE_1(Opcode):
    # Store a double into local variable 1
    # Stack: value →
    code = 0x48

    def __init__(self):
        super(DSTORE_1, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class DSTORE_2(Opcode):
    # Store a double into local variable 2
    # Stack: value →
    code = 0x49

    def __init__(self):
        super(DSTORE_2, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class DSTORE_3(Opcode):
    # Store a double into local variable 3
    # Stack: value →
    code = 0x4a

    def __init__(self):
        super(DSTORE_3, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class DSUB(Opcode):
    # Subtract a double from another
    # Stack: value1, value2 → result
    code = 0x67

    def __init__(self):
        super(DSUB, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class DUP(Opcode):
    # Duplicate the value on top of the stack
    # value → value, value
    code = 0x59

    def __init__(self):
        super(DUP, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class DUP_X1(Opcode):
    # Insert a copy of the top value into the stack two values from the top. value1
    # and value2 must not be of the type double or long.
    # Stack: value2, value1 → value1, value2, value1
    code = 0x5a

    def __init__(self):
        super(DUP_X1, self).__init__()

    @property
    def produce_count(self):
        return 3

    @property
    def consume_count(self):
        return 2


class DUP_X2(Opcode):
    # Insert a copy of the top value into the stack two (if value2 is double or long
    # it takes up the entry of value3, too) or three values (if value2 is neither
    # double nor long) from the top
    # Stack: value3, value2, value1 → value1, value3, value2, value1
    code = 0x5b

    def __init__(self):
        super(DUP_X2, self).__init__()

    @property
    def produce_count(self):
        return 4

    @property
    def consume_count(self):
        return 3


class DUP2(Opcode):
    # Duplicate top two stack words (two values, if value1 is not double nor long; a
    # single value, if value1 is double or long)
    # Stack: {value2, value1} → {value2, value1}, {value2, value1}
    code = 0x5c

    def __init__(self):
        super(DUP2, self).__init__()

    @property
    def produce_count(self):
        return 4

    @property
    def consume_count(self):
        return 2


class DUP2_X1(Opcode):
    code = 0x5d

    def __init__(self):
        super(DUP2_X1, self).__init__()
# value3, {value2, value1} → {value2, value1}, value3, {value2, value1}
# Duplicate two words and insert beneath third word (see explanation above)

    @property
    def produce_count(self):
        return 4

    @property
    def consume_count(self):
        return 3


class DUP2_X2(Opcode):
    # Duplicate two words and insert beneath fourth word
    # Stack: {value4, value3}, {value2, value1} → {value2, value1}, {value4, value3}, {value2, value1}
    code = 0x5e

    def __init__(self):
        super(DUP2_X2, self).__init__()

    @property
    def produce_count(self):
        return 6

    @property
    def consume_count(self):
        return 4


class F2D(Opcode):
    # Convert a float to a double
    # Stack: value → result
    code = 0x8d

    def __init__(self):
        super(F2D, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class F2I(Opcode):
    # Convert a float to an int
    # Stack: value → result
    code = 0x8b

    def __init__(self):
        super(F2I, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class F2L(Opcode):
    # Convert a float to a long
    # Stack: value → result
    code = 0x8c

    def __init__(self):
        super(F2L, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class FADD(Opcode):
    # Add two floats
    # value1, value2 → result
    code = 0x62

    def __init__(self):
        super(FADD, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class FALOAD(Opcode):
    # Load a float from an array
    # Stack: arrayref, index → value
    code = 0x30

    def __init__(self):
        super(FALOAD, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class FASTORE(Opcode):
    # Store a float in an array
    # Stack: arrayref, index, value →
    code = 0x51

    def __init__(self):
        super(FASTORE, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 3


class FCMPG(Opcode):
    # Compare two floats
    # Stack: value1, value2 → result
    code = 0x96

    def __init__(self):
        super(FCMPG, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class FCMPL(Opcode):
    # Compare two floats
    # Stack: value1, value2 → result
    code = 0x95

    def __init__(self):
        super(FCMPL, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class FCONST_0(Opcode):
    # Push 0.0f on the stack
    # Stack: → 0.0f
    code = 0x0b

    def __init__(self):
        super(FCONST_0, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class FCONST_1(Opcode):
    # Push 1.0f on the stack
    # Stack: → 1.0f
    code = 0x0c

    def __init__(self):
        super(FCONST_1, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class FCONST_2(Opcode):
    # Push 2.0f on the stack
    # Stack: → 2.0f
    code = 0x0d

    def __init__(self):
        super(FCONST_2, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class FDIV(Opcode):
    # Divide two floats
    # Stack: value1, value2 → result
    code = 0x6e

    def __init__(self):
        super(FDIV, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class FLOAD(Opcode):
    # Load a float value from a local variable #index
    # Args(1): index
    # Stack: → value
    code = 0x17

    def __init__(self, var):
        super(FLOAD, self).__init__()
        self.var = var

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.var

    @classmethod
    def read_extra(cls, reader, dump=None):
        var = reader.read_u1()
        return cls(var)

    def write_extra(self, writer):
        writer.write_u1(self.var)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class FLOAD_0(Opcode):
    # Load a float value from local variable 0
    # Stack: → value
    code = 0x22

    def __init__(self):
        super(FLOAD_0, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class FLOAD_1(Opcode):
    # Load a float value from local variable 1
    # Stack: → value
    code = 0x23

    def __init__(self):
        super(FLOAD_1, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class FLOAD_2(Opcode):
    # Load a float value from local variable 2
    # Stack: → value
    code = 0x24

    def __init__(self):
        super(FLOAD_2, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class FLOAD_3(Opcode):
    # Load a float value from local variable 3
    # Stack: → value
    code = 0x25

    def __init__(self):
        super(FLOAD_3, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class FMUL(Opcode):
    # Multiply two floats
    # Stack: value1, value2 → result
    code = 0x6a

    def __init__(self):
        super(FMUL, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class FNEG(Opcode):
    # Negate a float
    # Stack: value → result
    code = 0x76

    def __init__(self):
        super(FNEG, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class FREM(Opcode):
    # Get the remainder from a division between two floats
    # Stack: value1, value2 → result
    code = 0x72

    def __init__(self):
        super(FREM, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class FRETURN(Opcode):
    # Return a float
    # Stack: value → [empty]
    code = 0xae

    def __init__(self):
        super(FRETURN, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class FSTORE(Opcode):
    # Store a float value into a local variable #index
    # Args(1): index
    # Stack: value →
    code = 0x38

    def __init__(self, var):
        super(FSTORE, self).__init__()
        self.var = var

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.var

    @classmethod
    def read_extra(cls, reader, dump=None):
        var = reader.read_u1()
        return cls(var)

    def write_extra(self, writer):
        writer.write_u1(self.var)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class FSTORE_0(Opcode):
    # Store a float value into local variable 0
    # Stack: value →
    code = 0x43

    def __init__(self):
        super(FSTORE_0, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class FSTORE_1(Opcode):
    # Store a float value into local variable 1
    # Stack: value →
    code = 0x44

    def __init__(self):
        super(FSTORE_1, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class FSTORE_2(Opcode):
    # Store a float value into local variable 2
    # Stack: value →
    code = 0x45

    def __init__(self):
        super(FSTORE_2, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class FSTORE_3(Opcode):
    # Store a float value into local variable 3
    # Stack: value →
    code = 0x46

    def __init__(self):
        super(FSTORE_3, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class FSUB(Opcode):
    # Subtract two floats
    # Stack: value1, value2 → result
    code = 0x66

    def __init__(self):
        super(FSUB, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class GETFIELD(Opcode):
    # Get a field value of an object objectref, where the field is identified by
    # field reference in the constant pool index (index1 << 8 + index2)
    # Args(2): index1, index2
    # Stack: objectref → value
    code = 0xb4

    def __init__(self, class_name, field_name, descriptor):
        super(GETFIELD, self).__init__()
        self.field = Fieldref(class_name, field_name, descriptor)

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s.%s (%s)' % (self.field.class_name, self.field.name, self.field.name_and_type.descriptor)

    @classmethod
    def read_extra(cls, reader, dump=None):
        field = reader.constant_pool[reader.read_u2()]
        return cls(
            field.class_name,
            field.name,
            field.name_and_type.descriptor.bytes.decode('mutf-8')
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.field))

    def resolve(self, constant_pool):
        self.field.resolve(constant_pool)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class GETSTATIC(Opcode):
    # Get a static field value of a class, where the field is identified by field
    # reference in the constant pool index (index 1 << 8 + index2)
    # Args(2): index1, index2
    # Stack: → value
    code = 0xb2

    def __init__(self, class_name, field_name, descriptor):
        super(GETSTATIC, self).__init__()
        self.field = Fieldref(class_name, field_name, descriptor)

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s.%s (%s)' % (self.field.class_name, self.field.name, self.field.name_and_type.descriptor)

    @classmethod
    def read_extra(cls, reader, dump=None):
        field = reader.constant_pool[reader.read_u2()]
        return cls(
            field.class_name,
            field.name,
            field.name_and_type.descriptor.bytes.decode('mutf-8')
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.field))

    def resolve(self, constant_pool):
        self.field.resolve(constant_pool)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class GOTO(Opcode):
    # Goes to another instruction at branchoffset (signed short constructed from
    # unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: [no change]
    code = 0xa7

    def __init__(self, offset):
        super(GOTO, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 0


class GOTO_W(Opcode):
    # Goes to another instruction at branchoffset (signed int constructed from
    # unsigned bytes branchbyte1 << 24 + branchbyte2 << 16 + branchbyte3 << 8 +
    # branchbyte4)
    # Args(4): branchbyte1, branchbyte2, branchbyte3, branchbyte4
    # Stack: [no change]
    code = 0xc8

    def __init__(self, offset):
        super(GOTO_W, self).__init__()
        self.offset = offset

    def __len__(self):
        return 5

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s4()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s4(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 0


class I2B(Opcode):
    # Convert an int into a byte
    # Stack: value → result
    code = 0x91

    def __init__(self):
        super(I2B, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class I2C(Opcode):
    # Convert an int into a character
    # Stack: value → result
    code = 0x92

    def __init__(self):
        super(I2C, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class I2D(Opcode):
    # Convert an int into a double
    # Stack: value → result
    code = 0x87

    def __init__(self):
        super(I2D, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class I2F(Opcode):
    # Convert an int into a float
    # Stack: value → result
    code = 0x86

    def __init__(self):
        super(I2F, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class I2L(Opcode):
    # Convert an int into a long
    # Stack: value → result
    code = 0x85

    def __init__(self):
        super(I2L, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class I2S(Opcode):
    # Convert an int into a short
    # Stack: value → result
    code = 0x93

    def __init__(self):
        super(I2S, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class IADD(Opcode):
    # Add two ints
    # Stack: value1, value2 → result
    code = 0x60

    def __init__(self):
        super(IADD, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class IALOAD(Opcode):
    # Load an int from an array
    # Stack: arrayref, index → value
    code = 0x2e

    def __init__(self):
        super(IALOAD, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class IAND(Opcode):
    # Perform a bitwise and on two integers
    # Stack: value1, value2 → result
    code = 0x7e

    def __init__(self):
        super(IAND, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class IASTORE(Opcode):
    # Store an int into an array
    # Stack: arrayref, index, value →
    code = 0x4f

    def __init__(self):
        super(IASTORE, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 3


class ICONST_M1(Opcode):
    # Load the int value -1 onto the stack
    # Stack: → -1
    code = 0x02

    def __init__(self):
        super(ICONST_M1, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ICONST_0(Opcode):
    # Load the int value 0 onto the stack
    # Stack: → 0
    code = 0x03

    def __init__(self):
        super(ICONST_0, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ICONST_1(Opcode):
    # Load the int value 1 onto the stack
    # Stack: → 1
    code = 0x04

    def __init__(self):
        super(ICONST_1, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ICONST_2(Opcode):
    # Load the int value 2 onto the stack
    # Stack: → 2
    code = 0x05

    def __init__(self):
        super(ICONST_2, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ICONST_3(Opcode):
    # Load the int value 3 onto the stack
    # Stack: → 3
    code = 0x06

    def __init__(self):
        super(ICONST_3, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ICONST_4(Opcode):
    # Load the int value 4 onto the stack
    # Stack: → 4
    code = 0x07

    def __init__(self):
        super(ICONST_4, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ICONST_5(Opcode):
    # Load the int value 5 onto the stack
    # Stack: → 5
    code = 0x08

    def __init__(self):
        super(ICONST_5, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class IDIV(Opcode):
    # Divide two integers
    # Stack: value1, value2 → result
    code = 0x6c

    def __init__(self):
        super(IDIV, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class IF_ACMPEQ(Opcode):
    # If references are equal, branch to instruction at branchoffset (signed short
    # constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value1, value2 →
    code = 0xa5

    def __init__(self, offset):
        super(IF_ACMPEQ, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 2


class IF_ACMPNE(Opcode):
    # If references are not equal, branch to instruction at branchoffset (signed
    # short constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value1, value2 →
    code = 0xa6

    def __init__(self, offset):
        super(IF_ACMPNE, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 2


class IF_ICMPEQ(Opcode):
    # If ints are equal, branch to instruction at branchoffset (signed short
    # constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value1, value2 →
    code = 0x9f

    def __init__(self, offset):
        super(IF_ICMPEQ, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 2


class IF_ICMPGE(Opcode):
    # If value1 is greater than or equal to value2, branch to instruction at
    # Iranchoffset (signed short constructed from unsigned bytes branchbyte1 << 8 +
    # branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value1, value2 →
    code = 0xa2

    def __init__(self, offset):
        super(IF_ICMPGE, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 2


class IF_ICMPGT(Opcode):
    # If value1 is greater than value2, branch to instruction at branchoffset
    # (signed short constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value1, value2 →
    code = 0xa3

    def __init__(self, offset):
        super(IF_ICMPGT, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 2


class IF_ICMPLE(Opcode):
    # If value1 is less than or equal to value2, branch to instruction at
    # Iranchoffset (signed short constructed from unsigned bytes branchbyte1 << 8 +
    # branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value1, value2 →
    code = 0xa4

    def __init__(self, offset):
        super(IF_ICMPLE, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 2


class IF_ICMPLT(Opcode):
    # If value1 is less than value2, branch to instruction at branchoffset (signed
    # short constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value1, value2 →
    code = 0xa1

    def __init__(self, offset):
        super(IF_ICMPLT, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 2


class IF_ICMPNE(Opcode):
    # If ints are not equal, branch to instruction at branchoffset (signed short
    # constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value1, value2 →
    code = 0xa0

    def __init__(self, offset):
        super(IF_ICMPNE, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 2


class IFEQ(Opcode):
    # If value is 0, branch to instruction at branchoffset (signed short constructed
    # from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value →
    code = 0x99

    def __init__(self, offset):
        super(IFEQ, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class IFGE(Opcode):
    # If value is greater than or equal to 0, branch to instruction at branchoffset
    # (signed short constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value →
    code = 0x9c

    def __init__(self, offset):
        super(IFGE, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class IFGT(Opcode):
    # If value is greater than 0, branch to instruction at branchoffset (signed
    # short constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value →
    code = 0x9d

    def __init__(self, offset):
        super(IFGT, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class IFLE(Opcode):
    # If value is less than or equal to 0, branch to instruction at branchoffset
    # (signed short constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value →
    code = 0x9e

    def __init__(self, offset):
        super(IFLE, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class IFLT(Opcode):
    # If value is less than 0, branch to instruction at branchoffset (signed short
    # constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value →
    code = 0x9b

    def __init__(self, offset):
        super(IFLT, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class IFNE(Opcode):
    # If value is not 0, branch to instruction at branchoffset (signed short
    # constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value →
    code = 0x9a

    def __init__(self, offset):
        super(IFNE, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class IFNONNULL(Opcode):
    # If value is not null, branch to instruction at branchoffset (signed short
    # constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value →
    code = 0xc7

    def __init__(self, offset):
        super(IFNONNULL, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class IFNULL(Opcode):
    # If value is null, branch to instruction at branchoffset (signed short
    # constructed from unsigned bytes branchbyte1 << 8 + branchbyte2)
    # Args(2): branchbyte1, branchbyte2
    # Stack: value →
    code = 0xc6

    def __init__(self, offset):
        super(IFNULL, self).__init__()
        self.offset = offset

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.offset

    @classmethod
    def read_extra(cls, reader, dump=None):
        offset = reader.read_s2()
        return cls(offset)

    def write_extra(self, writer):
        writer.write_s2(self.offset)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class IINC(Opcode):
    # Increment local variable #index by signed byte const
    # Args(2): index, const
    # Stack: [No change]
    code = 0x84

    def __init__(self, index, const):
        super(IINC, self).__init__()
        self.index = index
        self.const = const

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s %s' % (self.index, self.const)

    @classmethod
    def read_extra(cls, reader, dump=None):
        index = reader.read_u1()
        const = reader.read_u1()
        return cls(index, const)

    def write_extra(self, writer):
        writer.write_u1(self.index)
        writer.write_u1(self.const)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 0


class ILOAD(Opcode):
    # Load an int value from a local variable #index
    # Args(1): index
    # Stack: → value
    code = 0x15

    def __init__(self, var):
        super(ILOAD, self).__init__()
        self.var = var

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.var

    @classmethod
    def read_extra(cls, reader, dump=None):
        var = reader.read_u1()
        return cls(var)

    def write_extra(self, writer):
        writer.write_u1(self.var)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ILOAD_0(Opcode):
    # Load a reference onto the stack from local variable 0
    # Stack: → value
    code = 0x1a

    def __init__(self):
        super(ILOAD_0, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ILOAD_1(Opcode):
    # Load a reference onto the stack from local variable 1
    # Stack: → value
    code = 0x1b

    def __init__(self):
        super(ILOAD_1, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ILOAD_2(Opcode):
    # Load a reference onto the stack from local variable 2
    # Stack: → value
    code = 0x1c

    def __init__(self):
        super(ILOAD_2, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class ILOAD_3(Opcode):
    # Load a reference onto the stack from local variable 3
    # Stack: → value
    code = 0x1d

    def __init__(self):
        super(ILOAD_3, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class IMPDEP1(Opcode):
    # Reserved for implementation-dependent operations within debuggers; should not
    # appear in any class file
    code = 0xfe

    def __init__(self):
        super(IMPDEP1, self).__init__()


class IMPDEP2(Opcode):
    # Reserved for implementation-dependent operations within debuggers; should not
    # appear in any class file
    code = 0xff

    def __init__(self):
        super(IMPDEP2, self).__init__()


class IMUL(Opcode):
    # Multiply two integers
    # Stack: value1, value2 → result
    code = 0x68

    def __init__(self):
        super(IMUL, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class INEG(Opcode):
    # Negate int
    # Stack: value → result
    code = 0x74

    def __init__(self):
        super(INEG, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class INSTANCEOF(Opcode):
    # Determines if an object objectref is of a given type, identified by class
    # reference index in constant pool (indexbyte1 << 8 + indexbyte2)
    # Args(2): indexbyte1, indexbyte2
    # Stack: objectref → result
    code = 0xc1

    def __init__(self, class_name):
        super(INSTANCEOF, self).__init__()
        self.klass = Classref(class_name)

    def __arg_repr__(self):
        return ' %s' % self.klass.class_name

    def __len__(self):
        return 3

    @classmethod
    def read_extra(cls, reader, dump=None):
        klass = reader.constant_pool[reader.read_u2()]
        return cls(
            klass.name.bytes.decode('mutf-8'),
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.klass))

    def resolve(self, constant_pool):
        self.klass.resolve(constant_pool)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


def _compose_descriptor(descriptor, args, returns):
    if descriptor is not None:
        if not (args is None and returns is None):
            raise TypeError("Cannot specify both a descriptor and args/returns")
        return descriptor

    return '(%s)%s' % (
        ''.join(a for a in args),
        returns
    )


class INVOKEDYNAMIC(Opcode):
    # Invokes a dynamic method and puts the result on the stack (might be void); the
    # method is identified by method reference index in constant pool (indexbyte1 <<
    # 8 + indexbyte2)
    # Args(4): indexbyte1, indexbyte2, 0, 0
    # Stack: [arg1, [arg2 ...]] → result
    code = 0xba

    def __init__(self, class_name, method_name, descriptor=None, args=None, returns=None):
        super(INVOKEDYNAMIC, self).__init__()
        descriptor = _compose_descriptor(descriptor, args, returns)
        self.method = Methodref(class_name, method_name, descriptor)

    def __arg_repr__(self):
        return ' %s.%s %s' % (self.method.class_name, self.method.name, self.method.name_and_type.descriptor)

    def __len__(self):
        return 5

    @classmethod
    def read_extra(cls, reader, dump=None):
        method = reader.constant_pool[reader.read_u2()]
        reader.read_u2()
        return cls(
            method.klass.name.bytes.decode('mutf-8'),
            method.name_and_type.name.bytes.decode('mutf-8'),
            method.name_and_type.descriptor.bytes.decode('mutf-8')
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.method))
        writer.write_u2(0)

    def resolve(self, constant_pool):
        self.method.resolve(constant_pool)

    @property
    def produce_count(self):
        return 0 if self.method.name_and_type.descriptor.bytes[-1] == ord('V') else 1

    @property
    def consume_count(self):
        return 1 + len(self.method.descriptor.parameters)


class INVOKEINTERFACE(Opcode):
    # Invokes an interface method on object objectref and puts the result on the
    # stack (might be void); the interface method is identified by method reference
    # index in constant pool (indexbyte1 << 8 + indexbyte2)
    # Args(4): indexbyte1, indexbyte2, count, 0
    # Stack: objectref, [arg1, arg2, ...] → result
    code = 0xb9

    def __init__(self, class_name, method_name, descriptor=None, args=None, returns=None):
        super(INVOKEINTERFACE, self).__init__()
        descriptor = _compose_descriptor(descriptor, args, returns)
        self.method = InterfaceMethodref(class_name, method_name, descriptor)

    def __arg_repr__(self):
        return ' %s.%s %s' % (self.method.class_name, self.method.name, self.method.name_and_type.descriptor)

    def __len__(self):
        return 5

    @classmethod
    def read_extra(cls, reader, dump=None):
        method = reader.constant_pool[reader.read_u2()]
        reader.read_u1()  # Count (can be interpolated from parameters)
        reader.read_u1()  # Blank value
        return cls(
            method.class_name,
            method.name,
            method.name_and_type.descriptor.bytes.decode('mutf-8'),
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.method))
        writer.write_u1(len(self.method.descriptor.parameters) + 1)
        writer.write_u1(0)

    def resolve(self, constant_pool):
        self.method.resolve(constant_pool)

    @property
    def produce_count(self):
        return 0 if self.method.name_and_type.descriptor.bytes[-1] == ord('V') else 1

    @property
    def consume_count(self):
        return 1 + len(self.method.descriptor.parameters)


class INVOKESPECIAL(Opcode):
    # Invoke instance method on object objectref and puts the result on the stack
    # (might be void); the method is identified by method reference index in
    # constant pool (indexbyte1 << 8 + indexbyte2)
    # Args (2): indexbyte1, indexbyte2
    # Stack: objectref, [arg1, arg2, ...] → result
    code = 0xb7

    def __init__(self, class_name, method_name, descriptor=None, args=None, returns=None):
        super(INVOKESPECIAL, self).__init__()
        descriptor = _compose_descriptor(descriptor, args, returns)
        self.method = Methodref(class_name, method_name, descriptor)

    def __arg_repr__(self):
        return ' %s.%s %s' % (
            self.method.klass.name,
            self.method.name_and_type.name,
            self.method.name_and_type.descriptor
        )

    def __len__(self):
        return 3

    @property
    def class_name(self):
        return self.method.class_name

    @property
    def method_name(self):
        return self.method.method_name

    @property
    def descriptor(self):
        return self.method.descriptor

    @classmethod
    def read_extra(cls, reader, dump=None):
        method = reader.constant_pool[reader.read_u2()]
        return cls(
            method.klass.name.bytes.decode('mutf-8'),
            method.name_and_type.name.bytes.decode('mutf-8'),
            method.name_and_type.descriptor.bytes.decode('mutf-8')
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.method))

    def resolve(self, constant_pool):
        self.method.resolve(constant_pool)

    @property
    def produce_count(self):
        return 0 if self.method.name_and_type.descriptor.bytes[-1] == ord('V') else 1

    @property
    def consume_count(self):
        return 1 + len(self.method.descriptor.parameters)


class INVOKESTATIC(Opcode):
    # Invoke a static method and puts the result on the stack (might be void); the
    # method is identified by method reference index in constant pool (indexbyte1 <<
    # 8 + indexbyte2)
    # Args(2): indexbyte1, indexbyte2
    # Stack: [arg1, arg2, ...] → result
    code = 0xb8

    def __init__(self, class_name, method_name, descriptor=None, args=None, returns=None):
        descriptor = _compose_descriptor(descriptor, args, returns)
        super(INVOKESTATIC, self).__init__()
        self.method = Methodref(class_name, method_name, descriptor)

    def __arg_repr__(self):
        return ' %s.%s %s' % (
            self.method.klass.name,
            self.method.name_and_type.name,
            self.method.name_and_type.descriptor
        )

    def __len__(self):
        return 3

    @classmethod
    def read_extra(cls, reader, dump=None):
        method = reader.constant_pool[reader.read_u2()]
        return cls(
            method.klass.name.bytes.decode('mutf-8'),
            method.name_and_type.name.bytes.decode('mutf-8'),
            method.name_and_type.descriptor.bytes.decode('mutf-8')
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.method))

    def resolve(self, constant_pool):
        self.method.resolve(constant_pool)

    @property
    def produce_count(self):
        return 0 if self.method.name_and_type.descriptor.bytes[-1] == ord('V') else 1

    @property
    def consume_count(self):
        return len(self.method.descriptor.parameters)


class INVOKEVIRTUAL(Opcode):
    # Invoke virtual method on object objectref and puts the result on the stack
    # (might be void); the method is identified by method reference index in
    # constant pool (indexbyte1 << 8 + indexbyte2)
    # Args(2): indexbyte1, indexbyte2
    # Stack: objectref, [arg1, arg2, ...] → result
    code = 0xb6

    def __init__(self, class_name, method_name, descriptor=None, args=None, returns=None):
        descriptor = _compose_descriptor(descriptor, args, returns)
        super(INVOKEVIRTUAL, self).__init__()
        self.method = Methodref(class_name, method_name, descriptor)

    def __arg_repr__(self):
        return ' %s.%s %s' % (
            self.method.klass.name,
            self.method.name_and_type.name,
            self.method.name_and_type.descriptor
        )

    def __len__(self):
        return 3

    @classmethod
    def read_extra(cls, reader, dump=None):
        method = reader.constant_pool[reader.read_u2()]
        return cls(
            method.klass.name.bytes.decode('mutf-8'),
            method.name_and_type.name.bytes.decode('mutf-8'),
            method.name_and_type.descriptor.bytes.decode('mutf-8')
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.method))

    def resolve(self, constant_pool):
        self.method.resolve(constant_pool)

    @property
    def produce_count(self):
        return 0 if self.method.name_and_type.descriptor.bytes[-1] == ord('V') else 1

    @property
    def consume_count(self):
        return 1 + len(self.method.descriptor.parameters)


class IOR(Opcode):
    # Bitwise int or
    # Stack: value1, value2 → result
    code = 0x80

    def __init__(self):
        super(IOR, self).__init__()


class IREM(Opcode):
    # Logical int remainder
    # Stack: value1, value2 → result
    code = 0x70

    def __init__(self):
        super(IREM, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class IRETURN(Opcode):
    # Return an integer from a method
    # Stack: value → [empty]
    code = 0xac

    def __init__(self):
        super(IRETURN, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ISHL(Opcode):
    # Int shift left
    # Stack: value1, value2 → result
    code = 0x78

    def __init__(self):
        super(ISHL, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class ISHR(Opcode):
    # Int arithmetic shift right
    # Stack: value1, value2 → result
    code = 0x7a

    def __init__(self):
        super(ISHR, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class ISTORE(Opcode):
    # Store a reference into a local variable #index
    # Args(1): index
    # Stack: objectref →
    code = 0x36

    def __init__(self, var):
        super(ISTORE, self).__init__()
        self.var = var

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.var

    @classmethod
    def read_extra(cls, reader, dump=None):
        var = reader.read_u1()
        return cls(var)

    def write_extra(self, writer):
        writer.write_u1(self.var)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ISTORE_0(Opcode):
    # Store int value into variable 0
    # Stack: value →
    code = 0x3b

    def __init__(self):
        super(ISTORE_0, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ISTORE_1(Opcode):
    # Store int value into variable 1
    # Stack: value →
    code = 0x3c

    def __init__(self):
        super(ISTORE_1, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ISTORE_2(Opcode):
    # Store int value into variable 2
    # Stack: value →
    code = 0x3d

    def __init__(self):
        super(ISTORE_2, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ISTORE_3(Opcode):
    # Store int value into variable 3
    # Stack: value →
    code = 0x3e

    def __init__(self):
        super(ISTORE_3, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class ISUB(Opcode):
    # Int subtract
    # Stack: value1, value2 → result
    code = 0x64

    def __init__(self):
        super(ISUB, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class IUSHR(Opcode):
    # Int logical shift right
    # Stack: value1, value2 → result
    code = 0x7c

    def __init__(self):
        super(IUSHR, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class IXOR(Opcode):
    # Int xor
    # Stack: value1, value2 → result
    code = 0x82

    def __init__(self):
        super(IXOR, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class JSR(Opcode):
    code = 0xa8

    def __init__(self):
        super(JSR, self).__init__()
# 2: branchbyte1, branchbyte2 → address
# Jump to subroutine at branchoffset (signed short constructed from unsigned
# bytes branchbyte1 << 8 + branchbyte2) and place the return address on the
# stack


class JSR_W(Opcode):
    code = 0xc9

    def __init__(self):
        super(JSR_W, self).__init__()
# 4: branchbyte1, branchbyte2, branchbyte3, branchbyte4   → address
# Jump to subroutine at branchoffset (signed int constructed from unsigned bytes
# branchbyte1 << 24 + branchbyte2 << 16 + branchbyte3 << 8 + branchbyte4) and
# place the return address on the stack


class L2D(Opcode):
    # Convert a long to a double
    # Stack: value → result
    code = 0x8a

    def __init__(self):
        super(L2D, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class L2F(Opcode):
    # Convert a long to a float
    # Stack: value → result
    code = 0x89

    def __init__(self):
        super(L2F, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class L2I(Opcode):
    # Convert a long to a int
    # Stack: value → result
    code = 0x88

    def __init__(self):
        super(L2I, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class LADD(Opcode):
    # add two longs
    # Stack: value1, value2 → result
    code = 0x61

    def __init__(self):
        super(LADD, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class LALOAD(Opcode):
    # Load a long from an array
    # Stack: arrayref, index → value
    code = 0x2f

    def __init__(self):
        super(LALOAD, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class LAND(Opcode):
    # bitwise and of two longs
    # Stack: value1, value2 → result
    code = 0x7f

    def __init__(self):
        super(LAND, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class LASTORE(Opcode):
    # Store a long to an array
    # Stack: arrayref, index, value →
    code = 0x50

    def __init__(self):
        super(LASTORE, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 3


class LCMP(Opcode):
    # Compare two longs values
    # Stack: value1, value2 → result
    code = 0x94

    def __init__(self):
        super(LCMP, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class LCONST_0(Opcode):
    # Push the long 0 onto the stack
    # Stack: → 0L
    code = 0x09

    def __init__(self):
        super(LCONST_0, self).__init__()

    @property
    def produce_count(self):
        # Although it only produces one value, the wide value
        # takes up 2 slots on the stack.
        return 2

    @property
    def consume_count(self):
        return 0


class LCONST_1(Opcode):
    # Push the long 1 onto the stack
    # Stack: → 1L
    code = 0x0a

    def __init__(self):
        super(LCONST_1, self).__init__()

    @property
    def produce_count(self):
        # Although it only produces one value, the wide value
        # takes up 2 slots on the stack.
        return 2

    @property
    def consume_count(self):
        return 0


class LDC(Opcode):
    # Push a constant #index from a constant pool (String, int or float) onto the
    # stack
    # Args(1): index
    # Stack: → value
    code = 0x12

    def __init__(self, const):
        super(LDC, self).__init__()
        if isinstance(const, str):
            self.const = String(const)
        elif isinstance(const, int):
            self.const = Integer(const)
        # elif isinstance(const, float):
        #     self.const = Float(const)
        elif isinstance(const, Constant):
            self.const = const
        else:
            raise TypeError('Invalid type for LDC: %s' % type(const))

    def __arg_repr__(self):
        return ' %s' % self.const

    def __len__(self):
        return 2

    @classmethod
    def read_extra(cls, reader, dump=None):
        const = reader.read_u1()
        return cls(reader.constant_pool[const])

    def write_extra(self, writer):
        writer.write_u1(writer.constant_pool.index(self.const))

    def resolve(self, constant_pool):
        self.const.resolve(constant_pool)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class LDC_W(Opcode):
    # Push a constant #index from a constant pool (String, int or float) onto the
    # stack (wide index is constructed as indexbyte1 << 8 + indexbyte2)
    # Args(2): indexbyte1, indexbyte2
    # Stack: → value
    code = 0x13

    def __init__(self, const):
        super(LDC_W, self).__init__()
        if isinstance(const, str):
            self.const = String(const)
        elif isinstance(const, int):
            self.const = Integer(const)
        elif isinstance(const, float):
            self.const = Float(const)
        elif isinstance(const, Constant):
            self.const = const
        else:
            raise TypeError('Invalid type for LDC_W: %s' % type(const))

    def __arg_repr__(self):
        return ' %s' % self.const

    def __len__(self):
        return 3

    @classmethod
    def read_extra(cls, reader, dump=None):
        const = reader.read_u2()
        return cls(reader.constant_pool[const])

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.const))

    def resolve(self, constant_pool):
        self.const.resolve(constant_pool)

    @property
    def produce_count(self):
        # Although it only produces one value, the wide value
        # takes up 2 slots on the stack.
        return 2

    @property
    def consume_count(self):
        return 0


class LDC2_W(Opcode):
    # Push a constant #index from a constant pool (double or long) onto the stack
    # (wide index is constructed as indexbyte1 << 8 + indexbyte2)
    # Args(2): indexbyte1, indexbyte2
    # Stack: → value
    code = 0x14

    def __init__(self, const):
        super(LDC2_W, self).__init__()
        if isinstance(const, float):
            self.const = Double(const)
        elif isinstance(const, int):
            self.const = Long(const)
        elif isinstance(const, Constant):
            self.const = const
        else:
            raise TypeError('Invalid type for LDC_W: %s' % type(const))

    def __arg_repr__(self):
        return ' %s' % self.const

    def __len__(self):
        return 3

    @classmethod
    def read_extra(cls, reader, dump=None):
        const = reader.read_u2()
        return cls(reader.constant_pool[const])

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.const))

    def resolve(self, constant_pool):
        self.const.resolve(constant_pool)

    @property
    def produce_count(self):
        # Although it only produces one value, the wide value
        # takes up 2 slots on the stack.
        return 2

    @property
    def consume_count(self):
        return 0


class LDIV(Opcode):
    # Divide two longs
    # Stack: value1, value2 → result
    code = 0x6d

    def __init__(self):
        super(LDIV, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class LLOAD(Opcode):
    # Load a long value from a local variable #index
    # Args(1): index
    # Stack: → value
    code = 0x16

    def __init__(self, var):
        super(LLOAD, self).__init__()
        self.var = var

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.var

    @classmethod
    def read_extra(cls, reader, dump=None):
        var = reader.read_u1()
        return cls(var)

    def write_extra(self, writer):
        writer.write_u1(self.var)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class LLOAD_0(Opcode):
    # Load a long value from a local variable 0
    # Stack: → value
    code = 0x1e

    def __init__(self):
        super(LLOAD_0, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class LLOAD_1(Opcode):
    # Load a long value from a local variable 1
    # Stack: → value
    code = 0x1f

    def __init__(self):
        super(LLOAD_1, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class LLOAD_2(Opcode):
    # Load a long value from a local variable 2
    # Stack: → value
    code = 0x20

    def __init__(self):
        super(LLOAD_2, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class LLOAD_3(Opcode):
    # Load a long value from a local variable 3
    # Stack: → value
    code = 0x21

    def __init__(self):
        super(LLOAD_3, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class LMUL(Opcode):
    # Multiply two longs
    # Stack: value1, value2 → result
    code = 0x69

    def __init__(self):
        super(LMUL, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class LNEG(Opcode):
    # Negate a long
    # Stack: value → result
    code = 0x75

    def __init__(self):
        super(LNEG, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class LOOKUPSWITCH(Opcode):
    code = 0xab

    def __init__(self):
        super(LOOKUPSWITCH, self).__init__()
# 4+: <0-3 bytes padding>, defaultbyte1, defaultbyte2, defaultbyte3, defaultbyte4,
#   npairs1, npairs2, npairs3, npairs4, match-offset pairs...
# key →
# A target address is looked up from a table using a key and execution continues
# from the instruction at that address


class LOR(Opcode):
    # Bitwise or of two longs
    # Stack: value1, value2 → result
    code = 0x81

    def __init__(self):
        super(LOR, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class LREM(Opcode):
    # Remainder of division of two longs
    # Stack: value1, value2 → result
    code = 0x71

    def __init__(self):
        super(LREM, self).__init__()

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 2


class LRETURN(Opcode):
    # Return a long value
    # Stack: value → [empty]
    code = 0xad

    def __init__(self):
        super(LRETURN, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class LSHL(Opcode):
    code = 0x79

    def __init__(self):
        super(LSHL, self).__init__()
# value1, value2 → result
# Bitwise shift left of a long value1 by int value2 positions


class LSHR(Opcode):
    code = 0x7b

    def __init__(self):

        super(LSHR, self).__init__()
# value1, value2 → result
# Bitwise shift right of a long value1 by int value2 positions


class LSTORE(Opcode):
    # Store a long value in a local variable #index
    # Args(1): index
    # Stack: value →
    code = 0x37

    def __init__(self):
        super(LSTORE, self).__init__()


class LSTORE_0(Opcode):
    # Store a long value in a local variable 0
    # Stack: value →
    code = 0x3f

    def __init__(self):
        super(LSTORE_0, self).__init__()


class LSTORE_1(Opcode):
    # Store a long value in a local variable 1
    # Stack: value →
    code = 0x40

    def __init__(self):
        super(LSTORE_1, self).__init__()


class LSTORE_2(Opcode):
    # Store a long value in a local variable 2
    # Stack: value →
    code = 0x41

    def __init__(self):
        super(LSTORE_2, self).__init__()


class LSTORE_3(Opcode):
    # Store a long value in a local variable 3
    # Stack: value →
    code = 0x42

    def __init__(self):
        super(LSTORE_3, self).__init__()


class LSUB(Opcode):
    code = 0x65

    def __init__(self):
        super(LSUB, self).__init__()
# value1, value2 → result subtract two longs


class LUSHR(Opcode):
    code = 0x7d

    def __init__(self):
        super(LUSHR, self).__init__()
# value1, value2 → result
# Bitwise shift right of a long value1 by int value2 positions, unsigned


class LXOR(Opcode):
    code = 0x83

    def __init__(self):
        super(LXOR, self).__init__()
# value1, value2 → result
# Bitwise exclusive or of two longs


class MONITORENTER(Opcode):
    code = 0xc2

    def __init__(self):
        super(MONITORENTER, self).__init__()
# objectref →
# Enter monitor for object ("grab the lock" - start of synchronized() section)


class MONITOREXIT(Opcode):
    code = 0xc3

    def __init__(self):
        super(MONITOREXIT, self).__init__()
# objectref →
# Exit monitor for object ("release the lock" - end of synchronized() section)


class MULTIANEWARRAY(Opcode):
    code = 0xc5

    def __init__(self):
        super(MULTIANEWARRAY, self).__init__()
# 3: indexbyte1, indexbyte2, dimensions
# count1, [count2,...] → arrayref
# Create a new array of dimensions dimensions with elements of type identified
# by class reference in constant pool index (indexbyte1 << 8 + indexbyte2); the
# sizes of each dimension is identified by count1, [count2, etc.]


class NEW(Opcode):
    # Create new object of type identified by class reference in constant pool index
    # (indexbyte1 << 8 + indexbyte2)
    # Args(2): indexbyte1, indexbyte2
    # Stack: → objectref
    code = 0xbb

    def __init__(self, class_name):
        super(NEW, self).__init__()
        self.classref = Classref(class_name)

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.classref.name

    @classmethod
    def read_extra(cls, reader, dump=None):
        classref = reader.constant_pool[reader.read_u2()]
        return cls(
            classref.name.bytes.decode('mutf-8'),
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.classref))

    def resolve(self, constant_pool):
        self.classref.resolve(constant_pool)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class NEWARRAY(Opcode):
    # Create new array with count elements of primitive type identified by atype
    # 1: atype
    # count → arrayref
    code = 0xbc

    # Valid primitive types that NEWARRAY can be created with.
    # For non-primitive types, see ANEWARRAY
    T_BOOLEAN = 4
    T_CHAR = 5
    T_FLOAT = 6
    T_DOUBLE = 7
    T_BYTE = 8
    T_SHORT = 9
    T_INT = 10
    T_LONG = 11

    def __init__(self, atype):
        super(NEWARRAY, self).__init__()
        self.atype = atype

    def __len__(self):
        return 2

    def __arg_repr__(self):
        return ' %s' % self.atype

    @classmethod
    def read_extra(cls, reader, dump=None):
        atype = reader.read_u1()
        return cls(atype)

    def write_extra(self, writer):
        writer.write_u1(self.atype)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 1


class NOP(Opcode):
    # Perform no operation
    # Stack: [No change]
    code = 0x00

    def __init__(self):
        super(NOP, self).__init__()


class POP(Opcode):
    # Discard the top value on the stack
    # Stack: value →
    code = 0x57

    def __init__(self):
        super(POP, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class POP2(Opcode):
    # Discard the top two values on the stack (or one value, if it is a double or long)
    # {value2, value1} →
    code = 0x58

    def __init__(self):
        super(POP2, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 2


class PUTFIELD(Opcode):
    # Set field to value in an object objectref, where the field is identified by a
    # field reference index in constant pool (indexbyte1 << 8 + indexbyte2)
    # Args(2): indexbyte1, indexbyte2
    # Stack: objectref, value →
    code = 0xb5

    def __init__(self, class_name, field_name, descriptor):
        super(PUTFIELD, self).__init__()
        self.field = Fieldref(class_name, field_name, descriptor)

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s.%s (%s)' % (self.field.klass.name, self.field.name, self.field.name_and_type.descriptor)

    @classmethod
    def read_extra(cls, reader, dump=None):
        field = reader.constant_pool[reader.read_u2()]
        return cls(
            field.class_name,
            field.name,
            field.name_and_type.descriptor.bytes.decode('mutf-8')
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.field))

    def resolve(self, constant_pool):
        self.field.resolve(constant_pool)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 2


class PUTSTATIC(Opcode):
    # Set static field to value in a class, where the field is identified by a field
    # reference index in constant pool (indexbyte1 << 8 + indexbyte2)
    # Args(2): indexbyte1, indexbyte2
    # Stack: value →
    code = 0xb3

    def __init__(self, class_name, field_name, descriptor):
        super(PUTSTATIC, self).__init__()
        self.field = Fieldref(class_name, field_name, descriptor)

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s.%s (%s)' % (self.field.klass.name, self.field.name, self.field.name_and_type.descriptor)

    @classmethod
    def read_extra(cls, reader, dump=None):
        field = reader.constant_pool[reader.read_u2()]
        return cls(
            field.class_name,
            field.name,
            field.name_and_type.descriptor.bytes.decode('mutf-8')
        )

    def write_extra(self, writer):
        writer.write_u2(writer.constant_pool.index(self.field))

    def resolve(self, constant_pool):
        self.field.resolve(constant_pool)

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 1


class RET(Opcode):
    code = 0xa9

    def __init__(self):
        super(RET, self).__init__()
# 1: index
# [No change]
# Continue execution from address taken from a local variable #index (the
# asymmetry with jsr is intentional)


class RETURN(Opcode):
    # Return void from method
    # Stack:→ [empty]
    code = 0xb1

    def __init__(self):
        super(RETURN, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 0


class SALOAD(Opcode):
    code = 0x35

    def __init__(self):
        super(SALOAD, self).__init__()
# arrayref, index → value
# Load short from array


class SASTORE(Opcode):
    code = 0x56

    def __init__(self):
        super(SASTORE, self).__init__()
# arrayref, index, value →
# Store short to array


class SIPUSH(Opcode):
    # push a short onto the stack
    # Args(2): byte1, byte2
    # Stack: → value
    code = 0x11

    def __init__(self, const):
        super(SIPUSH, self).__init__()
        self.const = const

    def __len__(self):
        return 3

    def __arg_repr__(self):
        return ' %s' % self.const

    @classmethod
    def read_extra(cls, reader, dump=None):
        const = reader.read_s2()
        return cls(const)

    def write_extra(self, writer):
        writer.write_s2(self.const)

    @property
    def produce_count(self):
        return 1

    @property
    def consume_count(self):
        return 0


class SWAP(Opcode):
    # Swaps two top words on the stack (note that value1 and value2 must not be
    # double or long)
    # Stack: value2, value1 → value1, value2
    code = 0x5f

    def __init__(self):
        super(SWAP, self).__init__()

    @property
    def produce_count(self):
        return 0

    @property
    def consume_count(self):
        return 0


class TABLESWITCH(Opcode):
    code = 0xaa

    def __init__(self):
        super(TABLESWITCH, self).__init__()
# 4+: [0-3 bytes padding], defaultbyte1, defaultbyte2, defaultbyte3, defaultbyte4,
#   lowbyte1, lowbyte2, lowbyte3, lowbyte4,
#   highbyte1, highbyte2, highbyte3, highbyte4, jump offsets...
# index →
# continue execution from an address in the table at offset index


class WIDE(Opcode):
    code = 0xc4

    def __init__(self):
        super(WIDE, self).__init__()
# 3/5: Opcode, indexbyte1, indexbyte2
# or
# iinc, indexbyte1, indexbyte2, countbyte1, countbyte2
# Execute Opcode, where Opcode is either iload, fload, aload, lload, dload,
# istore, fstore, astore, lstore, dstore, or ret, but assume the index is 16
# bit; or execute iinc, where the index is 16 bits and the constant to
# increment by is a signed 16 bit short



from ..utils import TranspileTestCase


class InterfaceTests(TranspileTestCase):

    def test_implement_interface(self):
        "You can implement (and use) a native Java interface"
        self.assertJavaExecution(
            """
            from java.lang import StringBuilder

            class MyStringAnalog(implements=java.lang.CharSequence):
                def __init__(self, value):
                    self.value = value

                def charAt(self, index: int) -> char:
                    return 'x'

                def length(self) -> int:
                    return len(self.value)

                def subSequence(self, start: int, end: int) -> java.lang.CharSequence:
                    return MyStringAnalog(self.value[start:end])

                def toString(self) -> java.lang.String:
                    return len(self.value) * 'x'

            analog = MyStringAnalog("world")

            builder = StringBuilder()

            builder.append("Hello, ")
            builder.append(analog)

            print(builder)
            print("Done.")
            """,
            """
            Hello, xxxxx
            Done.
            """, run_in_function=False)

    def test_implement_inner_interface(self):
        "You can implement (and use) a native Java interface defined as an inner class"
        self.assertJavaExecution(
            """
            from com.example import View

            class MyHandler(implements=com.example.View[Handler]):
                def event(self, view: com.example.View, name: java.lang.String) -> void:
                    print("My handler for a", name, "event on", view)

            view = View()

            handler = MyHandler()

            view.setHandler(handler)

            view.doEvent("Stuff")

            print("Done.")
            """,
            java={
                'com/example/View': """
                package com.example;

                public class View {
                    public static interface Handler {
                        public void event(View v, String event);
                    }

                    Handler handler;

                    public void setHandler(Handler h) {
                        System.out.println("Set handler to " + h);
                        this.handler = h;
                    }

                    public void doEvent(String name) {
                        System.out.println("Do event " + name);
                        this.handler.event(this, name);
                        System.out.println("Event done.");
                    }

                    public String toString() {
                        return "Java View";
                    }
                }
                """
            },
            out="""
            Set handler to <MyHandler object at 0xXXXXXXXX>
            Do event Stuff
            My handler for a Stuff event on Java View
            Event done.
            Done.
            """, run_in_function=False)



from ..utils import TranspileTestCase

from unittest import expectedFailure


class WithLoopTests(TranspileTestCase):
    def test_with(self):
        self.assertCodeExecution("""
            class CtxMgr:
                def __enter__(self):
                    print('entering CtxMgr')
                    return 42
                def __exit__(self, *args):
                    print('exiting CtxMgr', args)

            with CtxMgr() as val:
                print('val', val)

            print('val outside block', val)

            val = 16
            with CtxMgr():
                print('in another ctx, val now is', val)
            """)

    def test_with_body_fails(self):
        # TODO: add support for exc_type and traceback information
        self.assertCodeExecution("""
            class CtxMgr:
                def __enter__(self):
                    print('entering CtxMgr')
                def __exit__(self, exc_type, exc_value, traceback):
                    print('exiting CtxMgr')
                    # print('exc_type', exc_type)
                    print('exc_value', exc_value)
                    # print('traceback', traceback)

            with CtxMgr():
                raise KeyError('ola')
            """, exits_early=True)

    @expectedFailure
    def test_with_suppresses_exception(self):
        self.assertCodeExecution("""
            class CtxMgr:
                def __enter__(self):
                    print('entering CtxMgr')
                def __exit__(self, exc_type, exc_value, traceback):
                    print('exiting CtxMgr')
                    print('exc_value', exc_value)
                    return True

            with CtxMgr():
                raise KeyError('ola')

            print('Done')
        """)

    def test_with_noexit(self):
        self.assertCodeExecution("""
            class CtxMgrMissingExit:
                def __enter__(self):
                    print('entering CtxMgrMissingExit')
            with CtxMgrMissingExit():
                print('inside')
        """, exits_early=True)

    def test_with_noenter(self):
        self.assertCodeExecution("""
            class CtxMgrMissingEnter:
                def __exit__(self, *args):
                    print('exiting CtxMgrMissingEnter')
            with CtxMgrMissingEnter():
                print('inside')
        """, exits_early=True)

    def test_with_nested(self):
        self.assertCodeExecution("""
            class CtxMgr:
                def __enter__(self):
                    print('entering CtxMgr')
                    return 42
                def __exit__(self, *args):
                    print('exiting CtxMgr')

            class CtxMgr2:
                def __enter__(self):
                    print('entering CtxMgr2')
                    return 24
                def __exit__(self, *args):
                    print('exiting CtxMgr2')

            with CtxMgr() as val, CtxMgr2() as val2:
                print('val', val)
                print('val2', val2)
            """)

    def test_with_tuple_variables(self):
        self.assertCodeExecution("""
            class CtxMgr:
                def __enter__(self):
                    return 1,2
                def __exit__(self, *args):
                    print("exiting....")
            with CtxMgr() as (a,b):
                print('a', a, 'b', b)

            """)



from itertools import permutations
from .. utils import (
    BinaryOperationTestCase,
    InplaceOperationTestCase,
    SAMPLE_SUBSTITUTIONS,
    TranspileTestCase,
    UnaryOperationTestCase,
)


class ListTests(TranspileTestCase):
    def test_setattr(self):
        self.assertCodeExecution("""
            x = [1, 2, 3]
            try:
                x.attr = 42
            except AttributeError as err:
                print(err)
            """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = [1, 2, 3]
            try:
                print(x.attr)
            except AttributeError as err:
                print(err)
            """)

    def test_creation(self):
        # Empty list
        self.assertCodeExecution("""
            x = []
            print(x)
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(x)
            """)

    def test_getitem(self):
        # Simple positive index
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(x[2])
            """)

        # Simple negative index
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(x[-2])
            """)

        # Positive index out of range
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            try:
                print(x[10])
            except IndexError as err:
                print(err)
            """)

        # Negative index out of range
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            try:
                print(x[-10])
            except IndexError as err:
                print(err)
            """)

    def test_setitem(self):
        self.assertCodeExecution("""
            x = [1]
            x[0] = 5
            print(x[0])
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3]
            x[1] = "hello"
            x[2] = "there"
            print(x)
            """)

        # Out of bounds
        self.assertCodeExecution("""
            x = []
            try:
                x[0] = 5
            except IndexError as err:
                print(err)
            """)

        # Out of bounds (negative)
        self.assertCodeExecution("""
            x = [1]
            try:
                x[-2] = 5
            except IndexError as err:
                print(err)
            """)

    def test_append(self):
        # New list
        self.assertCodeExecution("""
            x = []
            x.append("hello")
            x.append(5)
            print(x[0], x[1])
            """)

        # Existing list
        self.assertCodeExecution("""
            x = [1, 2, 3, 4]
            x.append(5)
            x.append("hello")
            print(x[4], x[5])
            """)

    def extend_substitutions(self, prefix, suffix):
        possible_vals = []
        for perm in permutations(suffix):
            possible_vals.append(repr(prefix + list(perm)))
        return possible_vals

    def test_extend(self):
        # extend a list
        self.assertCodeExecution("""
            x = [1, 2, 3]
            x.extend([4, "hello"])
            print(x)
            """)

        # extend a tuple
        x = [1, 2, 3]
        y = [5, "world", "hello"]
        self.assertCodeExecution("""
            x = %(x)s
            x.extend(%(y)s)
            print(x)
            """ % {
                'x': x,
                'y': y,
            },
            substitutions={repr(x + y): self.extend_substitutions(x, y)})

        # extend a frozenset
        x = [1, 2, 3]
        y = [8, "theta"]
        self.assertCodeExecution("""
            x = %(x)s
            y = frozenset(%(y)s)
            x.extend(y)
            """ % {
                'x': x,
                'y': y,
            },
            substitutions={repr(x + y): self.extend_substitutions(x, y)})

        # extend a set
        x = [1, 2, 3]
        y = ["beta", "alpha"]
        self.assertCodeExecution("""
            x = %(x)s
            y = set(%(y)s)
            x.extend(y)
            """ % {
                'x': x,
                'y': y,
            },
            substitutions={repr(x + y): self.extend_substitutions(x, y)})

        # extend a dict
        x = [1, 2, 3]
        y = {"alpha": 4, "theta": 6, "beta": 5}
        keys = list(y.keys())
        self.assertCodeExecution("""
            x = %(x)s
            y = %(y)s
            x.extend(y)
            """ % {
                'x': x,
                'y': y,
            },
            substitutions={repr(x + keys): self.extend_substitutions(x, keys)})

        # extend an iterator
        self.assertCodeExecution("""
            x = [1, 2, 3]
            x.extend(iter([4,5,6]))
            print(x)
            """)

        # extend bytearray
        self.assertCodeExecution("""
            x = [1, 2, 3]
            x.extend(bytearray(b'abc'))
            print(x)
            """)

        # extend byte
        self.assertCodeExecution("""
            x = [1, 2, 3]
            x.extend(b'def')
            print(x)
            """)

        # extend str
        self.assertCodeExecution("""
            x = [1, 2, 3]
            x.extend(b'string')
            print(x)
            """)

        # extend range
        self.assertCodeExecution("""
            x = [1, 2, 3]
            x.extend(range(5))
            print(x)
            """)

        # extend non-iterable
        self.assertCodeExecution("""
            x = [1, 2, 3]
            try:
                x.extend(4)
            except TypeError as err:
                print(err)
            print(x)
            """)

    def test_remove(self):
        # Remove integer
        self.assertCodeExecution("""
            x = [1, 2, 3]
            x.remove(1)
            print(x)
            """)

        # Remove only first duplicate
        self.assertCodeExecution("""
            x = [1, 2, 2, 3, 2]
            x.remove(2)
            print(x)
            """)

        # Remove boolean
        self.assertCodeExecution("""
            x = [True, False, True, False]
            x.remove(1)
            print(x)
            """)

        # Not in list
        self.assertCodeExecution("""
            x = [1, 2]
            try:
                x.remove(3)
            except ValueError as err:
                print(err)
            print(x)
            """)

    def test_reverse(self):
        # New list
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            x.reverse()
            print(x)
            """)

    def test_slice(self):
        # Full slice
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(x[:])
            """)

        # Left bound slice
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(x[1:])
            """)

        # Right bound slice
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(x[:4])
            """)

        # Slice bound in both directions
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(x[1:4])
            """)

        # Slice bound in both directions with end out of bounds
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(x[1:6])
            """)

        # Slice bound in both directions with start out of bounds
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(x[6:7])
            """)

    # when step is 0
    def test_slice_with_zero_step(self):
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            try:
                print(x[1:3:0])
            except ValueError as err:
                print(err)
            """)

    def test_slice_in_reverse(self):
        # Full slice with a negative step
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print (x[::-1])
            """)

        # left bound slice with a negative step
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print (x[4::-2])
            """)

        # Right bound slice with a negative step
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print (x[:4:-1])
            """)

        # Right bound and left bound slice with a negative step
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print (x[1:4:-2])
            """)

    def test_count(self):
        # Normal Count
        self.assertCodeExecution("""
            x = [1, 1, 1, 4, 5]
            print(x.count(1))
            """)

        # Bool Count
        self.assertCodeExecution("""
            x = [1, 1, False, 1, 4, True, 5, True]
            print(x.count(1))
            """)

        # Element doesn't exist count
        self.assertCodeExecution("""
            x = [1, False, 1, 1, True, 4, 5, True]
            print(x.count(2))
            """)

        self.assertCodeExecution("""
            x = [1, 1, 1, 4, 5, True]
            print(x.count(1))
            """)

    def test_contains(self):
        # Normal Contains
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(1 in x)
            """)

        # Element doesn't exist
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            print(0 in x)
            """)

        # Checking for boolean
        self.assertCodeExecution("""
            x = [True, False]
            print(x.count(1))
            """)

    def test_sort(self):
        self.assertCodeExecution("""
            fixtures = [
                [9, 4, 7],
                ['beta', 'theta', 'alpha'],
            ]
            for x in fixtures:
                x.sort()
                print(x)
            """)

        self.assertCodeExecution("""
            fixtures = [
                [9, 4, 7],
                ['beta', 'theta', 'alpha'],
            ]
            for x in fixtures:
                x.sort(reverse=True)
                print(x)
            """)

        self.assertCodeExecution("""
            def second(s):
                return s[1]

            x = ['abc', 'bza', 'cda', 'daa']
            x.sort(key=second)
            print(x)
            """)

        self.assertCodeExecution("""
            def second(s):
                return s[1]

            x = ['abc', 'bza', 'cda', 'daa']
            x.sort(key=second, reverse=True)
            print(x)
            """)

    def test_pop(self):
        self.assertCodeExecution("""
            x = [1, 2, 3]
            print(x.pop())
            print(x)
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3]
            print(x.pop(0))
            print(x)
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3]
            print(x.pop(-2))
            print(x)
            """)

    def test_pop_exceptions(self):
        self.assertCodeExecution("""
            x = []
            try:
                print(x.pop())
            except IndexError as err:
                print(err)
            print(x)
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3]
            try:
                print(x.pop(3))
            except IndexError as err:
                print(err)
            print(x)
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3]
            try:
                print(x.pop(-4))
            except IndexError as err:
                print(err)
            print(x)
            """)

    def test_copy(self):
        self.assertCodeExecution("""
            x = [1, 2, 3]
            y = x.copy()
            print(y)
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3]
            y = x.copy()
            print(x == y)
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3]
            y = x.copy()
            print(x is not y)
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3]
            y = x.copy()
            y.append(4)
            print(x == y)
            """)

        self.assertCodeExecution("""
            x = [[1], 2, 3]
            y = x.copy()
            print(x[0] is y[0])
            """)

    def test_index(self):
        self.assertCodeExecution("""
            x = [1, 2, 3]
            print(x.index(1))
            """)

        self.assertCodeExecution("""
            x = [1, 2, 1]
            print(x.index(1, 1))
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3, 4]
            print(x.index(4, 0, len(x)))
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3, 4]
            print(x.index(2, 1, 2))
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3, 4]
            print(x.index(2, 0, 10))
            """)

        self.assertCodeExecution("""
            x = [1, 2, 1]
            print(x.index(1, 0, -2))
            """)

        self.assertCodeExecution("""
            x = [1, 2, 1]
            print(x.index(1, -3, -2))
            """)

        # cases for 'ValueError: not in list'
        self.assertCodeExecution("""
            x = [1, 2, 3]
            try:
                print(x.index(4))
            except ValueError as err:
                print(err)

            x = [1, 2, 1]
            try:
                print(x.index(2, 0, 1))
            except ValueError as err:
                print(err)

            x = [1, 2, 3, 4]
            try:
                print(x.index(4, 0, 3))
            except ValueError as err:
                print(err)

            x = [1, 2, 1]
            try:
                print(x.index(3, 0, 10))
            except ValueError as err:
                print(err)

            x = [1, 2, 3, 4]
            try:
                print(x.index(2, 10, 20))
            except ValueError as err:
                print(err)

            x = [1, 2, 3, 4]
            try:
                print(x.index(2, 10, 0))
            except ValueError as err:
                print(err)

            x = []
            try:
                print(x.index(1, 0, 10))
            except ValueError as err:
                print(err)
            """)

    def test_insert(self):
        self.assertCodeExecution("""
            x = [1, 2, 3]
            x.insert(0, 4)
            print(x)
            x.insert(1, 5)
            print(x)
            x.insert(len(x), 6)
            print(x)
            x.insert(200, 7)
            print(x)
            x.insert(-1, 8)
            print(x)
            x.insert(-len(x), 9)
            print(x)
            x.insert(-200, 10)
            print(x)
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3]
            x.insert(0, "hello")
            print(x)
            """)

        self.assertCodeExecution("""
            x = [1, 2, 3]
            x.insert(0, [1, 2])
            print(x)
            """)

        # cases for 'TypeError: wrong type'
        self.assertCodeExecution("""
            x = [1, 2, 3]
            try:
                print(x.insert('a', 4))
            except TypeError as err:
                print(err)

            x = [1, 2, 3]
            try:
                print(x.insert([1,2,3], 4))
            except TypeError as err:
                print(err)
            """)

    def test_lt_reflected(self):
        self.assertCodeExecution("""
            class A:
                def __gt__(self, other):
                    return True

            x = A()
            y = A()

            # verify that A doesn't have __lt__()
            print(x.__lt__(x))

            # ensure rich comparison logic is used
            print([x] < [x]) # False, x is x and same size
            print([x] < [y]) # True, x is not y, reflected

            # when elements are non-identical, return that comparison, even if size is not
            print([x, y] < [y]) # True, x is not y, reflected

            # ensure tie breaker by size is still used when identical elements
            print([x, y] < [x]) # False, larger size
            print([x] < [x, y]) # True, smaller size
            """)

    def test_le_reflected(self):
        self.assertCodeExecution("""
            class A:
                def __ge__(self, other):
                    return True

            x = A()
            y = A()

            # verify that A doesn't have __le__()
            print(x.__le__(x))

            # ensure rich comparison logic is used
            print([x] <= [x]) # False, x is x and same size
            print([x] <= [y]) # True, x is not y, reflected

            # when elements are non-identical, return that comparison, even if size is not
            print([x, y] <= [y]) # True, x is not y, reflected

            # ensure tie breaker by size is still used when identical elements
            print([x, y] <= [x]) # False, larger size
            print([x] <= [x, y]) # True, smaller size
            """)

    def test_gt_reflected(self):
        self.assertCodeExecution("""
            class A:
                def __lt__(self, other):
                    return True

            x = A()
            y = A()

            # verify that A doesn't have __gt__()
            print(x.__gt__(x))

            # ensure rich comparison logic is used
            print([x] > [x]) # False, x is x and same size
            print([x] > [y]) # True, x is not y, reflected

            # when elements are non-identical, return that comparison, even if size is not
            print([x, y] > [y]) # True, x is not y, reflected

            # ensure tie breaker by size is still used when identical elements
            print([x, y] > [x]) # False, larger size
            print([x] > [x, y]) # True, smaller size
            """)

    def test_ge_reflected(self):
        self.assertCodeExecution("""
            class A:
                def __le__(self, other):
                    return True

            x = A()
            y = A()

            # verify that A doesn't have __ge__()
            print(x.__ge__(x))

            # ensure rich comparison logic is used
            print([x] >= [x]) # False, x is x and same size
            print([x] >= [y]) # True, x is not y, reflected

            # when elements are non-identical, return that comparison, even if size is not
            print([x, y] >= [y]) # True, x is not y, reflected

            # ensure tie breaker by size is still used when identical elements
            print([x, y] >= [x]) # False, larger size
            print([x] >= [x, y]) # True, smaller size
            """)

    def test_eq_reflected(self):
        self.assertCodeExecution("""
            class A:
                def __eq__(self, other):
                    return True

            class B:
                def __eq__(self, other):
                    return False

            x = A()
            y = B()

            print([x] == [x]) # True, identity implies equality
            print([x, x] == [x]) # False, size not equal

            print([x] == [y]) # True, x is not y, x.__eq__(y)
            print([y] == [x]) # False, y is not x, y.__eq__(x)
            """)

    def test_reversed(self):
        self.assertCodeExecution("""
            list = [123, 'xyz', 'abc']
            reversedList = ['abc', 'xyz', 123] # to check equality
            l = []
            reverse = list.__reversed__()
            try:
                while(True):
                    l.append(next(reverse))
            except StopIteration:
                print(l == reversedList)

            origList = [123, 'xyz', 'abc'] # to check if original list was modified
            print(list == origList)
        """)

    def test_too_many_arguments(self):
        self.assertCodeExecution("""
            try:
                print(list(1, 2, 3))
            except TypeError as err:
                print(err)
        """)


class UnaryListOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'list'


class BinaryListOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'list'


class InplaceListOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'list'

    # Several tests add a list to an iterable with a non-deterministic
    # ordering. We define these substitutions here where the first part of the
    # values are the list (in its original order) and the last part is the
    # non-deterministic iterable.
    substitutions = {}
    for prefix in (
        [],
        ['a', 'b', 'c'],
        [3, 4, 5],
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    ):
        for suffix in (
            [1, 2.3456, 7],
            ['on', 'to', 'an'],
            ['one', 'two', 'six'],
            ['a', 'd', 'c'],
        ):
            from_vals = []
            for perm in permutations(suffix):
                from_vals.append(repr(prefix + list(perm)))
            substitutions[repr(prefix + suffix)] = from_vals

    substitutions.update(SAMPLE_SUBSTITUTIONS)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase, SAMPLE_SUBSTITUTIONS


class ListTests(TranspileTestCase):
    pass


class BuiltinListFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["list"]

    substitutions = dict(SAMPLE_SUBSTITUTIONS)
    substitutions.update({
        "[1, 2.3456, 'another']": [
            "[1, 'another', 2.3456]",
            "[2.3456, 1, 'another']",
            "[2.3456, 'another', 1]",
            "['another', 1, 2.3456]",
            "['another', 2.3456, 1]",
        ],
        "['a', 'c', 'd']": [
            "['a', 'd', 'c']",
            "['c', 'a', 'd']",
            "['c', 'd', 'a']",
            "['d', 'a', 'c']",
            "['d', 'c', 'a']",
        ]
    })



from .. utils import SAMPLE_DATA, TranspileTestCase, BuiltinFunctionTestCase


def _iterate_test(datatype):

    def test_func(self):
        code = '\n'.join([
            '\nfor x in {value}:\n    print(x)\n'.format(value=value)
            for value in SAMPLE_DATA[datatype]
        ])
        self.assertCodeExecution(code)

    return test_func


class ReversedTests(TranspileTestCase):
    test_iterate_bytearray = _iterate_test('bytearray')
    test_iterate_bytes = _iterate_test('bytes')
    test_iterate_list = _iterate_test('list')
    test_iterate_range = _iterate_test('range')
    test_iterate_str = _iterate_test('str')
    test_iterate_tuple = _iterate_test('tuple')


class BuiltinReversedFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["reversed"]



from ..utils import TranspileTestCase


class SetComprehensionTests(TranspileTestCase):
    def test_syntax(self):
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            s = {v**2 for v in x}
            print(len(s))
            print(1 in s)
            print(4 in s)
            print(9 in s)
            print(16 in s)
            print(25 in s)
            """)

    def test_method(self):
        self.assertCodeExecution("""
            x = [1, 2, 3, 4, 5]
            s = set(v**2 for v in x)
            print(len(s))
            print(1 in s)
            print(4 in s)
            print(9 in s)
            print(16 in s)
            print(25 in s)
            """)



r"""
This file defines a codec called "mutf-8", or Modified UTF-8. This is the codec
that Java uses as an on-the-wire format for JNI and class files.

The format is a combination of CESU-8, the encoding formed by layering UTF-8 on
top of UTF-16, as well as encoding NUL (codepoint 0) as a two byte sequence.
"""

from __future__ import unicode_literals

import codecs
import re
from encodings import normalize_encoding
from encodings.utf_8 import (
    IncrementalDecoder as UTF8IncrementalDecoder,
    IncrementalEncoder as UTF8IncrementalEncoder,
)

# Match 6-byte CESU-9 sequences
CESU8_RE = re.compile(b'\xed[\xa0-\xaf][\x80-\xbf]\xed[\xb0-\xbf][\x80-\xbf]')


class IncrementalDecoder(UTF8IncrementalDecoder):
    def _buffer_decode(self, input, errors, final):
        pieces = []
        position = 0
        while True:
            # Consume a chunk
            decoded, consumed = self._buffer_decode_chunk(
                input[position:],
                errors,
                final
            )
            # if we consumed nothing, it's probably the end of
            # the stream.
            if consumed == 0:
                break

            # Append the decoded text to the list, and update our position.
            pieces.append(decoded)
            position += consumed

        if final:
            # Check that we've consumed everything.
            assert position == len(input)

        return ''.join(pieces), position

    def _buffer_decode_chunk(self, input, errors, final):
        # CESU-8 sequences start with 0xed
        # Java NULLs start with 0xc0.
        index1 = input.find(b'\xed')
        index2 = input.find(b'\xc0')

        # Set `index` to whichever index comes first.
        if index1 != -1 and index2 != -1:
            index = min(index1, index2)
        elif index1 != -1:
            index = index1
        elif index2 != -1:
            index = index2
        else:
            # None of the marker characters were found, so
            # encode the entire input as normal UTF-8
            return super()._buffer_decode(input, errors, final)

        if index1 == 0:
            # Decode a six-byte CESU-8 sequence
            return self._buffer_decode_surrogates(input, errors, final)
        elif index2 == 0:
            # Decode a Java NULL
            return self._buffer_decode_null(input, errors, final)
        else:
            # Decode the bytes up until the next weird thing as UTF-8.
            # Set final=True because 0xc0 and 0xed don't make sense in the
            # middle of a sequence, in any variant.
            return super()._buffer_decode(input[:index], errors, True)

    def _buffer_decode_null(self, input, errors, final):
        """
        Decode the bytes 0xc0 0x80 as U+0000, like Java does.
        """
        nextbyte = input[1:2]
        if nextbyte == b'':
            if final:
                # We found 0xc0 at the end of the stream, which is an error.
                # Delegate to the superclass method to handle that error.
                return super()._buffer_decode(input, errors, final)
            else:
                # We found 0xc0 and we don't know what comes next, so consume
                # no bytes and wait.
                return '', 0
        elif nextbyte == b'\x80':
            # We found the usual 0xc0 0x80 sequence, so decode it and consume
            # two bytes.
            return '\u0000', 2
        else:
            # We found 0xc0 followed by something else, which is an error.
            # Whatever should happen is equivalent to what happens when the
            # superclass is given just the byte 0xc0, with final=True.
            return super()._buffer_decode(b'\xc0', errors, True)

    def _buffer_decode_surrogates(self, input, errors, final):
        """
        When we have improperly encoded surrogates, we can still see the
        bits that they were meant to represent.
        The surrogates were meant to encode a 20-bit number, to which we
        add 0x10000 to get a codepoint. That 20-bit number now appears in
        this form:
          11101101 1010abcd 10efghij 11101101 1011klmn 10opqrst
        The CESU8_RE above matches byte sequences of this form. Then we need
        to extract the bits and assemble a codepoint number from them.
        """
        if len(input) < 6:
            if final:
                # We found 0xed near the end of the stream, and there aren't
                # six bytes to decode. Delegate to the superclass method to
                # handle it as normal UTF-8. It might be a Hangul character
                # or an error.
                return super()._buffer_decode(input, errors, final)
            else:
                # We found 0xed, the stream isn't over yet, and we don't know
                # enough of the following bytes to decode anything, so consume
                # zero bytes and wait.
                return '', 0
        else:
            if CESU8_RE.match(input):
                # If this is a CESU-8 sequence, do some math to pull out
                # the intended 20-bit value, and consume six bytes.
                bytenums = input[:6]
                codepoint = (
                    ((bytenums[1] & 0x0f) << 16) +
                    ((bytenums[2] & 0x3f) << 10) +
                    ((bytenums[4] & 0x0f) << 6) +
                    (bytenums[5] & 0x3f) +
                    0x10000
                )
                return chr(codepoint), 6
            else:
                # This looked like a CESU-8 sequence, but it wasn't one.
                # 0xed indicates the start of a three-byte sequence, so give
                # three bytes to the superclass to decode as usual -- except
                # for working around the Python 2 discrepancy as before.
                return super()._buffer_decode(input[:3], errors, False)


class IncrementalEncoder(UTF8IncrementalEncoder):
    def encode(self, input, final=False):
        # encode input (taking the buffer into account)
        data = self.buffer + input
        (result, consumed) = self._buffer_encode(data, self.errors, final)
        # keep unencoded input until the next call
        self.buffer = data[consumed:]
        return result

    def _buffer_encode(self, input, errors, final):
        encoded_segments = []
        position = 0
        while True:
            encoded, consumed = self._buffer_encode_codepoint(
                input[position:],
                final
            )
            if consumed == 0:
                # Either there's nothing left to encode, or we need to wait
                # for more input. Either way, we're done for now.
                break

            # Append the encoded text to the list, and update our position.
            encoded_segments.append(encoded)
            position += consumed

        if final:
            # _buffer_decode_chunk must consume all the bytes when `final` is
            # true.
            assert position == len(input)

        return b''.join(encoded_segments), position

    def _buffer_encode_codepoint(self, input, final):
        # Find the next byte position that indicates a variant of UTF-8.
        # CESU-8 sequences always start with 0xed,
        # Java nulls always start with 0xc0.

        if len(input) == 0:
            return codecs.utf_8_encode("")
        if ord(input[0]) > 0xffff:
            return _to_cesu_bytearray(input)
        elif input[0] == '\x00':
            return _java_null_sequence()
        elif "\ud800" <= input[0] <= "\udfff":
            if "strict" in self.errors:
                return _encode_unicode_surrogate(input)
            else:
                return codecs.utf_8_encode(input[0], self.errors)
        index = _index_of_first_special_character(input)
        if index == -1:
            # No special characters found; encode the remaining
            # input as standard UTF-8
            return codecs.utf_8_encode(input, self.errors)
        else:
            # Encode the bytes up until the next weird thing as UTF-8.
            return codecs.utf_8_encode(input[:index], self.errors)


def _index_of_first_special_character(input):
    special_detectors = [
        lambda c: ord(c) > 0xffff,
        lambda c: c == '\x00',
        lambda c: "\ud800" <= c <= "\udfff"
        ]
    for ix, ch in enumerate(input):
        if any([check_fn(ch) for check_fn in special_detectors]):
            return ix
    return -1


def _encode_unicode_surrogate(input):
    encoded = bytes(bytearray([
        0xED,
        0xA0 | ((ord(input[0]) >> 6) & 0x0f),
        0x80 | (ord(input[0]) & 0x3f),
    ]))
    return encoded, 1


def _to_cesu_bytearray(input):
    # Encode a six-byte sequence starting with 0xed.
    return bytes(bytearray([
        0xED,
        0xA0 | (((ord(input[0]) >> 16) - 1) & 0x0f),
        0x80 | ((ord(input[0]) >> 10) & 0x3f),
        0xED,
        0xB0 | ((ord(input[0]) >> 6) & 0x0f),
        0x80 | (ord(input[0]) & 0x3f),
    ])), 1


def _java_null_sequence():
    # Encode a two-byte sequence, 0xc0 0x80.
    return b'\xc0\x80', 1


# Everything below here is boilerplate that matches the modules in the
# built-in `encodings` package.
def encode(input, errors='strict', final=True):
    return IncrementalEncoder(errors).encode(input, final=final), len(input)


def decode(input, errors='strict', final=True):
    return IncrementalDecoder(errors).decode(input, final=final), len(input)


class StreamWriter(codecs.StreamWriter):
    def __init__(self, stream, errors='strict', mapping=None):
        super().__init__(stream, errors)
        self.encoder = IncrementalEncoder(errors)

    def encode(self, input, errors='strict'):
        return self.encoder.encode(input, final=False), len(input)


class StreamReader(codecs.StreamReader):
    def __init__(self, stream, errors='strict', mapping=None):
        super().__init__(stream, errors)
        self.decoder = IncrementalDecoder(errors)

    def decode(self, input, errors='strict'):
        return self.decoder.decode(input, final=False), len(input)


# encodings module API

def search_function(encoding):
    if normalize_encoding(encoding) == 'mutf_8':
        return codecs.CodecInfo(
            name='mutf-8',
            encode=encode,
            decode=decode,
            incrementalencoder=IncrementalEncoder,
            incrementaldecoder=IncrementalDecoder,
            streamreader=StreamReader,
            streamwriter=StreamWriter,
        )
    else:
        return None



from .. utils import TranspileTestCase


class SetattrTests(TranspileTestCase):
    def test_minimal(self):
        self.assertCodeExecution("""
            class MyClass(object):
                class_value = 42

                def __init__(self, val):
                    self.value = val

            print("On class: ")
            setattr(MyClass, 'class_value', 37)
            setattr(MyClass, 'other_class_value', 42)
            print('  class_value =', MyClass.class_value)
            print('  other_class_value =', MyClass.other_class_value)

            obj = MyClass(37)

            print("On instance:")
            setattr(obj, 'value', 37)
            setattr(obj, 'other_value', 42)
            print('  value =', obj.value)
            print('  other_value =', obj.other_value)
            """, run_in_function=False)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class IssubclassTests(TranspileTestCase):
    def test_no_args(self):
        self.assertCodeExecution("""
            try:
                print(issubclass())
            except TypeError as err:
                print(err)
            """, run_in_function=False)

    def test_one_arg(self):
        self.assertCodeExecution("""
            try:
                print(issubclass(123))
            except TypeError as err:
                print(err)
            """, run_in_function=False)

    def test_non_class_first_arg(self):
        self.assertCodeExecution("""
            try:
                print(issubclass(0, 0))
            except TypeError as err:
                print(err)
            """, run_in_function=False)

    def test_non_tuple_second_arg(self):
        self.assertCodeExecution("""
            class MyClass1:
                pass

            try:
                print(issubclass(MyClass1, [int, float]))
            except TypeError as err:
                print(err)
            """, run_in_function=False)

    def test_single_class(self):
        self.assertCodeExecution("""
            class MyClass1:
                pass

            class MyClass2:
                pass

            print(issubclass(MyClass1, MyClass1))
            print(issubclass(MyClass1, MyClass2))
            """, run_in_function=False)

    def test_multiple_classes(self):
        self.assertCodeExecution("""
            class MyClass1:
                pass

            class MyClass2:
                pass

            print(issubclass(MyClass1, (MyClass1, MyClass2)))
            """, run_in_function=False)

    def test_parent(self):
        self.assertCodeExecution("""
            class MyClass1:
                pass

            class MyClass2(MyClass1):
                pass

            print(issubclass(MyClass2, MyClass1))
            print(issubclass(MyClass1, MyClass2))
            """, run_in_function=False)

    def test_grandparent(self):
        self.assertCodeExecution("""
            class MyClass1:
                pass

            class MyClass2(MyClass1):
                pass

            class MyClass3(MyClass2):
                pass

            print(issubclass(MyClass2, MyClass1))
            print(issubclass(MyClass3, MyClass2))
            print(issubclass(MyClass3, MyClass1))
            """, run_in_function=False)

    def test_builtins(self):
        self.assertCodeExecution("""
            builtin_classes = (object, bytearray, bytes, complex, dict, int,
                float, frozenset, list, memoryview, set, str, tuple, bool)
            for base in builtin_classes:
                for sub in builtin_classes:
                    print(sub, base, issubclass(sub, base))
            """, run_in_function=False)


class BuiltinIssubclassFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["issubclass"]

    not_implemented = [
    ]



from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class RangeTests(TranspileTestCase):
    def test_creation(self):
        self.assertCodeExecution("""
            x = range(0, 5)
            print("x[0] = ", x[0])
            print("x[1] = ", x[1])
            print("x[3] = ", x[3])
            print("x[-1] = ", x[-1])
            try:
                print("x[5] = ", x[5])
            except IndexError as err:
                print(err)
            """)

    def test_step(self):
        self.assertCodeExecution("""
            x = range(0, 5, 2)
            print("x[0] = ", x[0])
            print("x[1] = ", x[1])
            print("x[-1] = ", x[-1])
            print("x[-3] = ", x[-3])
            try:
                print("x[3] = ", x[3])
            except IndexError as err:
                print(err)
            try:
                print("x[5] = ", x[5])
            except IndexError as err:
                print(err)
            try:
                print("x[-4] = ", x[-4])
            except IndexError as err:
                print(err)

            y = range(7, 1, -3)
            print("y[0] = ", y[0])
            print("y[1] = ", y[1])
            print("y[-2] = ", y[-2])
            """)

    def test_zero_step(self):
        self.assertCodeExecution("""
            try:
                range(0, 5, 0)
            except ValueError as err:
                print(err)
        """)

    def test_len_empty(self):
        self.assertCodeExecution("""
            print(len(range(5, 5)))
        """)

    def test_len_positive_step(self):
        self.assertCodeExecution("""
            print(len(range(5, 0, 1)))
        """)

    def test_len_negative_step(self):
        self.assertCodeExecution("""
            print(len(range(0, 5, -1)))
        """)

    def test_multiple_iterators(self):
        self.assertCodeExecution("""
            r = range(0, 10)
            print(list(r))
            print(list(r))
        """)

    def test_iterator_iterator(self):
        self.assertCodeExecution("""
            r = range(0, 10)
            i = iter(r)
            print(next(i))
            print(next(iter(i)))
            print(r)
        """)

    def test_setitem(self):
        self.assertCodeExecution("""
            r = range(10)
            try:
                r[0] = "abc"
            except TypeError as e:
                print(e)
        """)


class UnaryRangeOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'range'


class BinaryRangeOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'range'


class InplaceRangeOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'range'



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class ComplexTests(TranspileTestCase):
    pass


class BuiltinComplexFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["complex"]

    not_implemented = [
        'test_complex',
    ]



from ..utils import TranspileTestCase


class ForLoopTests(TranspileTestCase):
    def test_for_over_range(self):
        # Normal range
        self.assertCodeExecution("""
            total = 0
            for i in range(0, 5):
                total = total + i
                print(i, total)
            """)

        # Empty range
        self.assertCodeExecution("""
            total = 0
            for i in range(0, 0):
                total = total + i
                print(i, total)
            """)

        # Stepped range
        self.assertCodeExecution("""
            total = 0
            for i in range(0, 10, 2):
                total = total + i
                print(i, total)
            """)

        # Reverse range
        self.assertCodeExecution("""
            total = 0
            for i in range(5, 0, -1):
                total = total + i
                print(i, total)
            """)

        # Test expression statement
        self.assertCodeExecution("""
            for i in range(10):
                [1, 2, 3]
                "abc" == "def"
            """)

    def test_for_over_iterable(self):
        self.assertCodeExecution("""
            total = 0
            for i in [1, 2, 3, 5]:
                total = total + i
                print(i, total)
            """)

        self.assertCodeExecution("""
            total = 0
            for i in []:
                total = total + i
                print(i, total)
            """)

    def test_for_else(self):
        self.assertCodeExecution("""
            total = 0
            for i in []:
                total = total + i
            else:
                total = -999
            print(total)
            """)

    def test_for_else_break(self):
        self.assertCodeExecution("""
            total = 0
            for i in []:
                total = total + i
                break
            else:
                total = -999
            print(total)
            """)

    def test_break(self):
        self.assertCodeExecution("""
            for i in range(1, 10):
                print(i, i % 5)
                if i % 5 == 0:
                    break
                print("after")
            """)

    def test_continue(self):
        self.assertCodeExecution("""
            for i in range(1, 10):
                print(i, i % 5)
                if i % 5 == 0:
                    continue
                print("after")
            """)

    def test_nested(self):
        self.assertCodeExecution("""
            i = 1
            j = 10
            for i in range(1, 10):
                for k in range(0, i):
                    print(i, j)
            """)

    def test_multiple_values(self):
        self.assertCodeExecution("""
            values = [
                (1, 2, 3),
                (4, 5, 6),
                (7, 8, 9),
            ]
            for x, y, z, in values:
                print(x, y, z)
            """)

    def test_multiple_values_iterator(self):
        self.assertCodeExecution("""
            values = {
                'a': 1,
                'b': 2,
                'c': 3,
            }
            for k, v in sorted(values.items()):
                print(k, v)
            print('And once again...')
            for item in sorted(values.items()):
                print(item[0], item[1])
            """)

    def test_recursive(self):
        self.assertCodeExecution("""
            def process(data):
                print('process: ', data)
                for datum in data:
                    process(datum)
                print('data processed: ', data)

            data = [[], [[], [], []], [[]]]

            process(data)
            """)



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class BinTests(TranspileTestCase):
    def test_int_but_no_index(self):
        self.assertCodeExecution("""
            class IntLike:
                def __init__(self, val):
                    self.val = val
                def __int__(self):
                    return self.val

            x = IntLike(5)
            try:
                print(bin(x))
            except TypeError as err:
                print(err)
            """)


class BuiltinBinFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["bin"]

    not_implemented = [
    ]



import re
from collections import namedtuple

# From: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html

##########################################################################
# 4.3. Descriptors and Signatures
##########################################################################

# A descriptor is a string representing the type of a field or method.
# Descriptors are represented in the class file format using modified UTF-8
# strings (§4.4.7) and thus may be drawn, where not further constrained, from
# the entire Unicode codespace.

# A signature is a string representing the generic type of a field or method, or
# generic type information for a class declaration.

# ------------------------------------------------------------------------
# 4.3.1. Grammar Notation
# ------------------------------------------------------------------------

# Descriptors and signatures are specified using a grammar. This grammar is a
# set of productions that describe how sequences of characters can form
# syntactically correct descriptors of various types. Terminal symbols of the
# grammar are shown in bold fixed-width font. Nonterminal symbols are shown in
# italic type. The definition of a nonterminal is introduced by the name of the
# nonterminal being defined, followed by a colon. One or more alternative right-
# hand sides for the nonterminal then follow on succeeding lines. For example,
# the production:


# FieldType:
#     BaseType
#     ObjectType
#     ArrayType

# states that a FieldType may represent either a BaseType, an ObjectType or an
# ArrayType.

# A nonterminal symbol on the right-hand side of a production that is followed
# by an asterisk (*) represents zero or more possibly different values produced
# from that nonterminal, appended without any intervening space. Similarly, a
# nonterminal symbol on the right-hand side of a production that is followed by
# an plus sign (+) represents one or more possibly different values produced
# from that nonterminal, appended without any intervening space. The production:


# MethodDescriptor:
#     ( ParameterDescriptor* ) ReturnDescriptor

# states that a MethodDescriptor represents a left parenthesis, followed by zero
# or more ParameterDescriptor values, followed by a right parenthesis, followed
# by a ReturnDescriptor.

MethodDescriptor = namedtuple('MethodDescriptor', ['parameters', 'return_type'])

# ------------------------------------------------------------------------
# 4.3.2. Field Descriptors
# ------------------------------------------------------------------------

# A field descriptor represents the type of a class, instance, or local
# variable. It is a series of characters generated by the grammar:


# FieldDescriptor:
#     FieldType

# FieldType:
#     BaseType
#     ObjectType
#     ArrayType

# BaseType:
#     B
#     C
#     D
#     F
#     I
#     J
#     S
#     Z

# ObjectType:
#     L ClassName ;

# ArrayType:
#     [ ComponentType

# ComponentType:
#     FieldType

# The characters of BaseType, the L and ; of ObjectType, and the [ of ArrayType
# are all ASCII characters.

# The ClassName represents a binary class or interface name encoded in internal
# form (§4.2.1).

# The interpretation of field descriptors as types is as shown in Table 4.2.

# A field descriptor representing an array type is valid only if it represents a
# type with 255 or fewer dimensions.

# Table 4.2. Interpretation of FieldType characters

# BaseType Character  Type    Interpretation
# B   byte    signed byte
# C   char    Unicode character code point in the Basic Multilingual Plane, encoded with UTF-16
# D   double  double-precision floating-point value
# F   float   single-precision floating-point value
# I   int integer
# J   long    long integer
# L ClassName ;   reference   an instance of class ClassName
# S   short   signed short
# Z   boolean true or false
# [   reference   one array dimension
# The field descriptor of an instance variable of type int is simply I.

# The field descriptor of an instance variable of type Object is
# Ljava/lang/Object;. Note that the internal form of the binary name for class
# Object is used.

# The field descriptor of an instance variable that is a multidimensional double
# array, double d[][][], is [[[D.

FIELD_DESCRIPTOR_RE = re.compile("\[*L[^;]+;|\[*[ZBCSIFDJV]")

# # ------------------------------------------------------------------------
# # 4.3.3. Method Descriptors
# # ------------------------------------------------------------------------

# A method descriptor represents the parameters that the method takes and the
# value that it returns:


# MethodDescriptor:
#     ( ParameterDescriptor* ) ReturnDescriptor

# A parameter descriptor represents a parameter passed to a method:


# ParameterDescriptor:
#     FieldType

# A return descriptor represents the type of the value returned from a method.
# It is a series of characters generated by the grammar:

# ReturnDescriptor:
#     FieldType
#     VoidDescriptor

# VoidDescriptor:
#     V
# The character V indicates that the method returns no value (its return type is void).

# A method descriptor is valid only if it represents method parameters with a
# total length of 255 or less, where that length includes the contribution for
# this in the case of instance or interface method invocations. The total length
# is calculated by summing the contributions of the individual parameters, where
# a parameter of type long or double contributes two units to the length and a
# parameter of any other type contributes one unit.

# The method descriptor for the method:

# Object m(int i, double d, Thread t) {..}

# is (IDLjava/lang/Thread;)Ljava/lang/Object;. Note that the internal forms of
# the binary names of Thread and Object are used.

# The method descriptor for m is the same whether m is a class method or an
# instance method. Although an instance method is passed this, a reference to
# the current class instance, in addition to its intended parameters, that fact
# is not reflected in the method descriptor. The reference to this is passed
# implicitly by the method invocation instructions of the Java Virtual Machine
# that invoke instance methods (§2.6.1). A reference to this is not passed to a
# class method.

def method_descriptor(descriptor):
    matches = FIELD_DESCRIPTOR_RE.findall(descriptor)
    return MethodDescriptor(parameters=matches[:-1], return_type=matches[-1])

# ------------------------------------------------------------------------
# 4.3.4. Signatures
# ------------------------------------------------------------------------

# Signatures are used to encode Java programming language type information that
# is not part of the Java Virtual Machine type system, such as generic type and
# method declarations and parameterized types. See The Java Language
# Specification, Java SE 7 Edition for details about such types.

# This kind of type information is needed to support reflection and debugging,
# and by a Java compiler.

# In the following, the terminal symbol Identifier is used to denote the name of
# a type, field, local variable, parameter, method, or type variable, as
# generated by a Java compiler. Such a name must not contain any of the ASCII
# characters . ; [ / < > : (that is, the characters forbidden in method names
# (§4.2.2) and also colon) but may contain characters that must not appear in an
# identifier in the Java programming language (JLS §3.8).

# A class signature, defined by the production ClassSignature, is used to encode
# type information about a class declaration. It describes any formal type
# parameters the class might have, and lists its (possibly parameterized) direct
# superclass and direct superinterfaces, if any.


# ClassSignature:
#     FormalTypeParametersopt SuperclassSignature SuperinterfaceSignature*

# A formal type parameter is described by its name, followed by its class and
# interface bounds. If the class bound does not specify a type, it is taken to
# be Object.


# FormalTypeParameters:
#     < FormalTypeParameter+ >

# FormalTypeParameter:
#     Identifier ClassBound InterfaceBound*

# ClassBound:
#     : FieldTypeSignatureopt

# InterfaceBound:
#     : FieldTypeSignature

# SuperclassSignature:
#     ClassTypeSignature

# SuperinterfaceSignature:
#     ClassTypeSignature

# A field type signature, defined by the production FieldTypeSignature, encodes
# the (possibly parameterized) type for a field, parameter or local variable.


# FieldTypeSignature:
#     ClassTypeSignature
#     ArrayTypeSignature
#     TypeVariableSignature

# A class type signature gives complete type information for a class or
# interface type. The class type signature must be formulated such that it can
# be reliably mapped to the binary name of the class it denotes by erasing any
# type arguments and converting each . character in the signature to a $
# character.


# ClassTypeSignature:
#     L PackageSpecifieropt SimpleClassTypeSignature ClassTypeSignatureSuffix* ;

# PackageSpecifier:
#     Identifier / PackageSpecifier*

# SimpleClassTypeSignature:
#     Identifier TypeArgumentsopt

# ClassTypeSignatureSuffix:
#     . SimpleClassTypeSignature

# TypeVariableSignature:
#     T Identifier ;

# TypeArguments:
#     < TypeArgument+ >

# TypeArgument:
#     WildcardIndicatoropt FieldTypeSignature
#     *

# WildcardIndicator:
#     +
#     -

# ArrayTypeSignature:
#     [ TypeSignature

# TypeSignature:
#     FieldTypeSignature
#     BaseType

# A method signature, defined by the production MethodTypeSignature, encodes the
# (possibly parameterized) types of the method's formal arguments and of the
# exceptions it has declared in its throws clause, its (possibly parameterized)
# return type, and any formal type parameters in the method declaration.


# MethodTypeSignature:
#     FormalTypeParametersopt (TypeSignature*) ReturnType ThrowsSignature*

# ReturnType:
#     TypeSignature
#     VoidDescriptor

# ThrowsSignature:
#     ^ ClassTypeSignature
#     ^ TypeVariableSignature

# If the throws clause of a method or constructor does not involve type
# variables, the ThowsSignature may be elided from the MethodTypeSignature.

# A Java compiler must output generic signature information for any class,
# interface, constructor or member whose generic signature in the Java
# programming language would include references to type variables or
# parameterized types.

# The signature and descriptor (§4.3.3) of a given method or constructor may not
# correspond exactly, due to compiler-generated artifacts. In particular, the
# number of TypeSignatures that encode formal arguments in MethodTypeSignature
# may be less than the number of ParameterDescriptors in MethodDescriptor.

# Oracle's Java Virtual Machine implementation does not check the well-
# formedness of the signatures described in this subsection during loading or
# linking. Instead, these checks are deferred until the signatures are used by
# reflective methods, as specified in the API of Class and members of
# java.lang.reflect. Future versions of a Java Virtual Machine implementation
# may be required to perform some or all of these checks during loading or
# linking.



from .. utils import TranspileTestCase, UnaryOperationTestCase, BinaryOperationTestCase, InplaceOperationTestCase


class BoolTests(TranspileTestCase):

    def test_setattr(self):
        self.assertCodeExecution("""
            x = True
            try:
                x.attr = 42
            except AttributeError as err:
                print(err)
            """)

    def test_getattr(self):
        self.assertCodeExecution("""
            x = True
            try:
                print(x.attr)
            except AttributeError as err:
                print(err)
            """)

    def test_setitem(self):
        self.assertCodeExecution("""
            x = True
            try:
                x[0] = 1
            except TypeError as err:
                print(err)
        """)

    def test_too_many_arguments(self):
        self.assertCodeExecution("""
            try:
                print(bool(1, 2))
            except TypeError as err:
                print(err)
        """)


class UnaryBoolOperationTests(UnaryOperationTestCase, TranspileTestCase):
    data_type = 'bool'


class BinaryBoolOperationTests(BinaryOperationTestCase, TranspileTestCase):
    data_type = 'bool'


class InplaceBoolOperationTests(InplaceOperationTestCase, TranspileTestCase):
    data_type = 'bool'



import unittest

from voc.transpiler import Transpiler
from voc.python.blocks import BlockCodeTooLarge
from voc.python.methods import MethodCodeTooLarge


class LargeCodeTest(unittest.TestCase):

    def test_block_large_code(self):
        "Test exception for large code of an anonymous block."
        large_code = 'print(1 + 2)\n' * 3000
        transpiler = Transpiler(verbosity=0)
        self.assertRaises(BlockCodeTooLarge, transpiler.transpile_string,
                          "test.py", large_code)

    def test_method_large_code(self):
        "Test exception for large code of a method."
        large_code = 'def test():\n' + '    print(1 + 2)\n' * 3000
        transpiler = Transpiler(verbosity=0)
        self.assertRaises(MethodCodeTooLarge, transpiler.transpile_string,
                          "test.py", large_code)
